from pathlib import Path

from meltano.core.plugin import ProjectPlugin, PluginRef, PluginType
from meltano.core.error import PluginInstallError
from meltano.core.plugin.error import PluginMissingError
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.setting_definition import SettingDefinition
from meltano.core.transform_add_service import TransformAddService
from meltano.core.behavior.hookable import hook


class DbtInvoker(PluginInvoker):
    def Popen_options(self):
        return {**super().Popen_options(), "cwd": self.plugin_config["project_dir"]}


class DbtPlugin(ProjectPlugin):
    __plugin_type__ = PluginType.TRANSFORMERS
    __invoker_cls__ = DbtInvoker

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)


class DbtTransformPluginInstaller:
    def __init__(self, project, plugin: ProjectPlugin):
        self.project = project
        self.plugin = plugin

    def install(self, reason):
        if reason in (PluginInstallReason.ADD, PluginInstallReason.UPGRADE):
            try:
                transform_add_service = TransformAddService(self.project)

                # Add repo to my-test-project/transform/packages.yml
                packages_path = transform_add_service.packages_file.relative_to(
                    self.project.root
                )
                print(
                    f"Adding dbt package '{self.plugin.pip_url}' to '{packages_path}'..."
                )
                transform_add_service.add_to_packages(self.plugin)

                # Add model and vars to my-test-project/transform/dbt_project.yml
                dbt_project_path = transform_add_service.dbt_project_file.relative_to(
                    self.project.root
                )
                print(f"Adding dbt model to '{dbt_project_path}'...")
                transform_add_service.update_dbt_project(self.plugin)
            except PluginMissingError:
                raise PluginInstallError(
                    "Transformer 'dbt' is not installed. Run `meltano add transformer dbt` to add it to your project."
                )
            except FileNotFoundError as err:
                relative_path = Path(err.filename).relative_to(self.project.root)
                raise PluginInstallError(
                    f"File {relative_path}' could not be found. Run `meltano add files dbt` to set up a dbt project."
                )
        else:
            print(
                f"Run `meltano add transform {self.plugin.name}` to re-add this transform to your dbt project."
            )


class DbtTransformPlugin(ProjectPlugin):
    __plugin_type__ = PluginType.TRANSFORMS
    __installer_cls__ = DbtTransformPluginInstaller

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    def is_invokable(self):
        return False

    @property
    def runner(self):
        return PluginRef(PluginType.TRANSFORMERS, "dbt")

    @property
    def extra_settings(self):
        return [
            SettingDefinition(
                name="_package_name", value="$MELTANO_TRANSFORM_NAMESPACE"
            ),
            SettingDefinition(name="_vars", kind="object", value={}),
        ]



import sqlalchemy
from flask import url_for

from .dashboards_helper import DashboardsHelper
from .reports_helper import ReportsHelper
from meltano.api.models.embed_token import EmbedToken, ResourceType
from meltano.core.m5o.dashboards_service import DashboardsService
from meltano.core.m5o.reports_service import ReportsService
from meltano.core.project import Project


class InvalidEmbedToken(Exception):
    """Occurs when an embed token isn't found."""

    def __init__(self, token):
        self.token = token


class EmbedsHelper:
    def generate_embed_snippet(self, session, resource_id, resource_type, today=None):
        try:
            embed_token = (
                session.query(EmbedToken).filter_by(resource_id=resource_id).one()
            )
            is_new = False
        except sqlalchemy.orm.exc.NoResultFound:
            embed_token = EmbedToken(
                resource_id=resource_id, resource_type=resource_type
            )
            session.add(embed_token)
            is_new = True
        finally:
            session.commit()

        resource_type = ResourceType(embed_token.resource_type)
        if resource_type is ResourceType.REPORT:
            min_width = "500px"
            min_height = "400px"
        elif resource_type is ResourceType.DASHBOARD:
            min_width = "100%"
            min_height = "100vh"
        inline_styles = f"margin: 0; padding: 0; border: 0; font-size: 100%; font: inherit; vertical-align: baseline; min-width: {min_width}; min-height: {min_height};"
        embed_url = url_for(
            "root.embed", token=embed_token.token, today=today, _external=True
        )
        snippet = f'<iframe src="{embed_url}" style="{inline_styles}" />'

        return {
            "is_new": is_new,
            "url": embed_url,
            "resource_type": embed_token.resource_type,
            "snippet": snippet,
        }

    def get_embed_from_token(self, session, token, today=None):
        try:
            embed_token = session.query(EmbedToken).filter_by(token=token).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise InvalidEmbedToken(token)

        resource_type = ResourceType(embed_token.resource_type)
        resource_id = embed_token.resource_id

        if resource_type is ResourceType.REPORT:
            resource_payload = self.get_report_resource(resource_id, today=today)
        elif resource_type is ResourceType.DASHBOARD:
            resource_payload = self.get_dashboard_resource(resource_id, today=today)

        return {
            "resource": resource_payload,
            "resource_type": embed_token.resource_type,
        }

    def get_dashboard_resource(self, resource_id, today=None):
        project = Project.find()
        reports_service = ReportsService(project)
        dashboards_service = DashboardsService(project)

        dashboard = dashboards_service.get_dashboard(resource_id)
        reports = [
            reports_service.get_report(report_id)
            for report_id in dashboard["report_ids"]
        ]
        dashboards_helper = DashboardsHelper()
        reports_with_query_results = dashboards_helper.get_dashboard_reports_with_query_results(
            reports, today=today
        )
        return {
            "dashboard": dashboard,
            "reports_with_query_results": reports_with_query_results,
        }

    def get_report_resource(self, resource_id, today=None):
        project = Project.find()
        reports_service = ReportsService(project)

        report = reports_service.get_report(resource_id)
        reports_helper = ReportsHelper()
        return reports_helper.get_report_with_query_results(report, today=today)



import os
import threading
import subprocess

from meltano.core.project import Project
from meltano.core.meltano_invoker import MeltanoInvoker
from meltano.core.utils.pidfile import PIDFile


class APIWorker(threading.Thread):
    def __init__(self, project: Project, reload=False):
        super().__init__()

        self.project = project
        self.reload = reload
        self.pid_file = PIDFile(self.project.run_dir("gunicorn.pid"))

    def run(self):
        args = ["--config", "python:meltano.api.wsgi", "--pid", str(self.pid_file)]

        if self.reload:
            args += ["--reload"]

        args += ["meltano.api.app:create_app()"]

        MeltanoInvoker(self.project).invoke(args, command="gunicorn")

    def pid_path(self):
        return self.project.run_dir(f"gunicorn.pid")

    def stop(self):
        self.pid_file.process.terminate()



from meltano.core.plugin import PluginRef
from werkzeug.routing import PathConverter


class PluginRefConverter(PathConverter):
    def to_url(self, value):
        return "/".join([value.type, value.name])

    def to_python(self, value):
        return PluginRef(*value.split("/"))



import os
import click
import fnmatch
import json
import logging

from . import cli
from .params import project
from .utils import CliError
from meltano.core.config_service import ConfigService
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin.singer.catalog import parse_select_pattern, SelectionType
from meltano.core.select_service import SelectService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine


def selection_color(selection):
    if selection is SelectionType.SELECTED:
        return "bright_green"
    elif selection is SelectionType.AUTOMATIC:
        return "bright_white"
    elif selection is SelectionType.EXCLUDED:
        return "red"


def selection_mark(selection):
    """
    Returns the mark to indicate the selection type of an attribute.

    Examples:
      [automatic]
      [selected ]
      [excluded ]
    """

    colwidth = max(map(len, SelectionType))  # size of the longest mark
    return f"[{selection:<{colwidth}}]"


@cli.command()
@click.argument("extractor")
@click.argument("entities_filter", default="*")
@click.argument("attributes_filter", default="*")
@click.option("--list", is_flag=True)
@click.option("--all", is_flag=True)
@click.option("--exclude", is_flag=True)
@project(migrate=True)
def select(project, extractor, entities_filter, attributes_filter, **flags):
    try:
        if flags["list"]:
            show(project, extractor, show_all=flags["all"])
        else:
            add(
                project,
                extractor,
                entities_filter,
                attributes_filter,
                exclude=flags["exclude"],
            )

        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_select(
            extractor=extractor,
            entities_filter=entities_filter,
            attributes_filter=attributes_filter,
            flags=flags,
        )
    except PluginExecutionError as err:
        raise CliError(f"Cannot list the selected attributes: {err}") from err


def add(project, extractor, entities_filter, attributes_filter, exclude=False):
    select_service = SelectService(project, extractor)
    select_service.select(entities_filter, attributes_filter, exclude)


def show(project, extractor, show_all=False):
    _, Session = project_engine(project)
    select_service = SelectService(project, extractor)

    session = Session()
    try:
        list_all = select_service.list_all(session)
    finally:
        session.close()

    # legend
    click.secho("Legend:")
    for sel_type in SelectionType:
        click.secho(f"\t{sel_type}", fg=selection_color(sel_type))

    # report
    click.secho("\nEnabled patterns:")
    for select in map(parse_select_pattern, select_service.current_select):
        color = "red" if select.negated else "white"
        click.secho(f"\t{select.raw}", fg=color)

    click.secho("\nSelected attributes:")
    for stream, prop in (
        (stream, prop)
        for stream in list_all.streams
        for prop in list_all.properties[stream.key]
    ):
        entry_selection = stream.selection + prop.selection
        mark = selection_mark(entry_selection)
        if show_all or entry_selection:
            click.secho(f"\t{mark} ", fg=selection_color(entry_selection), nl=False)
            click.secho(stream.key, fg=selection_color(stream.selection), nl=False)
            click.secho(f".{prop.key}", fg=selection_color(entry_selection))



import click
from typing import List

from meltano.core.project_add_service import (
    ProjectAddService,
    PluginNotSupportedException,
    PluginAlreadyAddedException,
)
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin_install_service import (
    PluginInstallService,
    PluginInstallReason,
)
from meltano.core.plugin import PluginType, VariantNotFoundError
from meltano.core.project import Project
from meltano.core.tracking import GoogleAnalyticsTracker


class CliError(Exception):
    pass


def print_added_plugin(project, plugin, plugin_def=None, related=False):
    descriptor = plugin.type.descriptor
    if related:
        descriptor = f"related {descriptor}"

    if plugin.should_add_to_file(project):
        click.secho(
            f"Added {descriptor} '{plugin.name}' to your Meltano project", fg="green"
        )
    else:
        click.secho(
            f"Adding {descriptor} '{plugin.name}' to your Meltano project...",
            fg="green",
        )

    if plugin_def:
        repo_url = plugin_def.repo
        if repo_url:
            click.echo(f"Repository:\t{repo_url}")

        docs_url = plugin_def.docs
        if docs_url:
            click.echo(f"Documentation:\t{docs_url}")


def add_plugin(
    project: Project,
    plugin_type: PluginType,
    plugin_name: str,
    add_service: ProjectAddService,
    variant=None,
):
    try:
        plugin = add_service.add(plugin_type, plugin_name, variant=variant)
        plugin_def = add_service.discovery_service.get_definition(plugin)

        print_added_plugin(project, plugin, plugin_def)
    except PluginAlreadyAddedException as err:
        click.secho(
            f"{plugin_type.descriptor.capitalize()} '{plugin_name}' is already in your Meltano project",
            fg="yellow",
            err=True,
        )
        plugin = err.plugin
    except VariantNotFoundError as err:
        raise CliError(str(err)) from err
    except (PluginNotSupportedException, PluginNotFoundError) as err:
        raise CliError(
            f"{plugin_type.descriptor.capitalize()} '{plugin_name}' is not known to Meltano"
        ) from err

    click.echo()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type=plugin_type, plugin_name=plugin_name)

    return plugin


def add_related_plugins(
    project, plugins, add_service: ProjectAddService, plugin_types=list(PluginType)
):
    added_plugins = []
    for plugin_install in plugins:
        related_plugins = add_service.add_related(
            plugin_install, plugin_types=plugin_types
        )
        for related_plugin in related_plugins:
            related_plugin_def = add_service.discovery_service.get_definition(
                related_plugin
            )
            print_added_plugin(
                project, related_plugin, related_plugin_def, related=True
            )
            click.echo()

        added_plugins.extend(related_plugins)

    return added_plugins


def install_status_update(data, reason):
    plugin = data["plugin"]

    if data["status"] == "running":
        verb = "Updating" if reason == PluginInstallReason.UPGRADE else "Installing"
        click.secho(f"{verb} {plugin.type.descriptor} '{plugin.name}'...")
    elif data["status"] == "error":
        click.secho(data["message"], fg="red")
        click.secho(data["details"], err=True)
    elif data["status"] == "warning":
        click.secho(f"Warning! {data['message']}.", fg="yellow")
    elif data["status"] == "success":
        verb = "Updated" if reason == PluginInstallReason.UPGRADE else "Installed"
        click.secho(f"{verb} {plugin.type.descriptor} '{plugin.name}'", fg="green")
        click.echo()


def install_plugins(project, plugins, reason=PluginInstallReason.INSTALL):
    install_service = PluginInstallService(project)
    install_status = install_service.install_plugins(
        plugins, status_cb=install_status_update, reason=reason
    )
    num_installed = len(install_status["installed"])
    num_failed = len(install_status["errors"])

    fg = "green"
    if num_failed >= 0 and num_installed == 0:
        fg = "red"
    elif num_failed > 0 and num_installed > 0:
        fg = "yellow"

    if len(plugins) > 1:
        verb = "Updated" if reason == PluginInstallReason.UPGRADE else "Installed"
        click.secho(f"{verb} {num_installed}/{num_installed+num_failed} plugins", fg=fg)

    return num_failed == 0



import os
import yaml
import json
import click
import sys
import logging
from urllib.parse import urlparse
from typing import List

from . import cli
from .params import project
from .utils import CliError, add_plugin, add_related_plugins, install_plugins
from meltano.core.plugin import PluginType
from meltano.core.config_service import ConfigService
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.project_add_service import ProjectAddService
from meltano.core.project_add_custom_service import ProjectAddCustomService
from meltano.core.plugin_install_service import PluginInstallReason


@cli.command()
@click.argument("plugin_type", type=click.Choice(PluginType.cli_arguments()))
@click.argument("plugin_name", nargs=-1, required=True)
@click.option("--variant")
@click.option("--custom", is_flag=True)
@click.option("--include-related", is_flag=True)
@project()
@click.pass_context
def add(ctx, project, plugin_type, plugin_name, variant=None, **flags):
    plugin_type = PluginType.from_cli_argument(plugin_type)
    plugin_names = plugin_name  # nargs=-1

    config_service = ConfigService(project)
    discovery_service = PluginDiscoveryService(project, config_service=config_service)

    if flags["custom"]:
        if plugin_type in (
            PluginType.TRANSFORMERS,
            PluginType.TRANSFORMS,
            PluginType.ORCHESTRATORS,
        ):
            raise CliError(f"--custom is not supported for {ctx.invoked_subcommand}")

        add_service_class = ProjectAddCustomService
    else:
        add_service_class = ProjectAddService

    add_service = add_service_class(
        project,
        plugin_discovery_service=discovery_service,
        config_service=config_service,
    )

    plugins = [
        add_plugin(
            project, plugin_type, plugin_name, variant=variant, add_service=add_service
        )
        for plugin_name in plugin_names
    ]

    related_plugin_types = [PluginType.FILES]
    if flags["include_related"]:
        related_plugin_types = list(PluginType)

    related_plugins = add_related_plugins(
        project, plugins, add_service=add_service, plugin_types=related_plugin_types
    )
    plugins.extend(related_plugins)

    # We will install the plugins in reverse order, since dependencies
    # are listed after their dependents in `related_plugins`, but should
    # be installed first.
    plugins.reverse()

    success = install_plugins(project, plugins, reason=PluginInstallReason.ADD)

    if not success:
        raise CliError("Failed to install plugin(s)")

    printed_empty_line = False
    for plugin in plugins:
        plugin_def = discovery_service.get_definition(plugin)
        docs_url = plugin_def.docs or plugin_def.repo
        if not docs_url:
            continue

        if not printed_empty_line:
            click.echo()
            printed_empty_line = True

        click.echo(
            f"To learn more about {plugin.type.descriptor} '{plugin.name}', visit {docs_url}"
        )



import logging

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler, EVENT_TYPE_MODIFIED
from meltano.core.project import Project
from meltano.core.compiler.project_compiler import ProjectCompiler


class CompileEventHandler(PatternMatchingEventHandler):
    def __init__(self, compiler):
        self.compiler = compiler

        super().__init__(ignore_patterns=["*.m5oc"])

    def on_any_event(self, event):
        try:
            self.compiler.compile()
        except Exception as e:
            logging.error(f"Compilation failed: {str(e)}")


class MeltanoCompilerWorker:
    def __init__(self, project: Project, compiler: ProjectCompiler = None):
        self.project = project
        self.compiler = compiler or ProjectCompiler(project)
        self.observer = self.setup_observer()

    @property
    def model_dir(self):
        return self.project.root_dir("model")

    def setup_observer(self):
        event_handler = CompileEventHandler(self.compiler)
        observer = Observer()
        observer.schedule(event_handler, str(self.model_dir), recursive=True)

        return observer

    def start(self):
        try:
            self.observer.start()
            logging.info(f"Auto-compiling models in '{self.model_dir}'")
        except OSError:
            # most probably INotify being full
            logging.warning(
                f"Model auto-compilation is disabled: INotify limit reached."
            )

    def stop(self):
        self.observer.stop()



import subprocess
import platform
import sys
import shutil
import os
import logging
from pathlib import Path
from collections import namedtuple

from .project import Project
from .error import SubprocessError


logger = logging.getLogger(__name__)

VenvSpecs = namedtuple("VenvSpecs", ("lib_dir", "bin_dir", "site_packages_dir"))

POSIX = VenvSpecs(
    lib_dir="lib",
    bin_dir="bin",
    site_packages_dir=os.path.join("lib", f"python{sys.version[:3]}", "site-packages"),
)

NT = VenvSpecs(
    lib_dir="Lib",
    bin_dir="Scripts",
    site_packages_dir=os.path.join("Lib", "site-packages"),
)


class VirtualEnv:
    PLATFORM_SPECS = {"Linux": POSIX, "Darwin": POSIX, "Windows": NT}

    def __init__(self, root: Path):
        self.root = root
        self._specs = self.__class__.platform_specs()

    @classmethod
    def platform_specs(cls):
        try:
            system = platform.system()
            return cls.PLATFORM_SPECS[system]
        except KeyError:
            raise Exception(f"Platform {system} not supported.")

    def __getattr__(self, attr):
        spec = getattr(self._specs, attr)
        return self.root.joinpath(spec)

    def __str__(self):
        return str(self.root)


class VenvService:
    def __init__(self, project):
        self.project = project

    def clean(self, namespace="", name=""):
        venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        try:
            shutil.rmtree(str(venv))
            logger.debug("Removed old virtual environment")
        except FileNotFoundError:
            # If the VirtualEnv has never been created before do nothing
            logger.debug("No old virtual environment to remove")
            pass

        return

    def create(self, namespace="", name=""):
        venv = VirtualEnv(self.project.venvs_dir(namespace, name))

        logger.debug("Creating virtual environment")
        venv_run = subprocess.run(
            [sys.executable, "-m", "venv", str(venv)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if venv_run.returncode != 0:
            raise SubprocessError(
                f"Could not create of the virtualenv for '{namespace}/{name}'", venv_run
            )

        python_path = venv.bin_dir.joinpath("python")

        upgrade_run = subprocess.run(
            [
                str(python_path),
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                "setuptools",
                "wheel",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if upgrade_run.returncode != 0:
            raise SubprocessError(
                f"Failed to upgrade pip to the latest version.", upgrade_run
            )

        return venv_run

    def install(self, pip_url, namespace="", name=""):
        """
        Install a package using `pip` in the proper virtual environment.

        This method is not threadsafe.
        """

        venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        python_path = venv.bin_dir.joinpath("python")

        meltano_pth_path = venv.site_packages_dir.joinpath("meltano_venv.pth")
        if meltano_pth_path.exists():
            os.remove(meltano_pth_path)

        logger.debug("Installing into new virtual environment")
        run = subprocess.run(
            [str(python_path), "-m", "pip", "install", *pip_url.split(" ")],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise SubprocessError(f"Failed to install plugin '{name}'.", run)

        return run

    def exec_path(self, bin, name=None, namespace=""):
        name = name or bin
        venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        return venv.bin_dir.joinpath(bin)



from meltano.core.plugin import ProjectPlugin, PluginType


class ModelPlugin(ProjectPlugin):
    __plugin_type__ = PluginType.MODELS

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)



import os
import sys
import click
import logging

from meltano.core.logging import setup_logging
from meltano.core.project import ProjectReadonly
from .cli import cli
from .utils import CliError
from . import (
    elt,
    schema,
    discovery,
    initialize,
    add,
    install,
    invoke,
    ui,
    upgrade,
    schedule,
    select,
    repl,
    config,
    model,
    user,
)

setup_logging()

logger = logging.getLogger(__name__)


def main():
    # mark the current process as executed via the `cli`
    os.environ["MELTANO_JOB_TRIGGER"] = os.getenv("MELTANO_JOB_TRIGGER", "cli")
    try:
        try:
            cli(obj={"project": None})
        except ProjectReadonly as err:
            raise CliError(
                f"The requested action could not be completed: {err}"
            ) from err
    except CliError as err:
        logger.debug(str(err), exc_info=True)
        click.secho(str(err), fg="red")
        sys.exit(1)



import os
import json

from os.path import join
from pathlib import Path
from meltano.core.project import Project


class SettingsHelper:
    def __init__(self):
        project = Project.find()
        self.settings_file_path = project.root_dir("model", "database.settings.m5o")
        if not self.settings_file_path.is_file():
            with open(self.settings_file_path, "w") as f:
                settings = {"settings": {"connections": []}}
                json.dump(settings, f)

    def get_connections(self):
        with open(self.settings_file_path) as f:
            settings_file = json.load(f)
        return settings_file

    def save_connection(self, connection):
        with open(self.settings_file_path, "r") as f:
            settings = json.load(f)
        settings["settings"]["connections"].append(connection)
        with open(self.settings_file_path, "w") as f:
            json.dump(settings, f)
        return settings

    def delete_connection(self, connection):
        with open(self.settings_file_path, "r") as f:
            settings = json.load(f)
        connections = settings["settings"]["connections"]
        updated_connections = [
            conn for conn in connections if conn["name"] != connection["name"]
        ]
        settings["settings"]["connections"] = updated_connections
        with open(self.settings_file_path, "w") as f:
            json.dump(settings, f)
        return settings



from flask import jsonify, request
from .dashboards_helper import DashboardsHelper
from meltano.core.project import Project
from meltano.core.m5o.dashboards_service import (
    DashboardAlreadyExistsError,
    DashboardDoesNotExistError,
    DashboardsService,
)
from meltano.core.schedule_service import ScheduleNotFoundError
from .errors import InvalidFileNameError
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import block_if_readonly
from flask_security import roles_required


dashboardsBP = APIBlueprint("dashboards", __name__)


def dashboards_service():
    project = Project.find()
    return DashboardsService(project)


@dashboardsBP.errorhandler(DashboardAlreadyExistsError)
def _handle(ex):
    dashboard_name = ex.dashboard["name"]
    return (
        jsonify(
            {
                "error": True,
                "code": f"A dashboard with the name '{dashboard_name}' already exists. Try renaming the dashboard.",
            }
        ),
        409,
    )


@dashboardsBP.errorhandler(DashboardDoesNotExistError)
def _handle(ex):
    dashboard_name = ex.dashboard["name"]
    return (
        jsonify(
            {"error": True, "code": f"The dashboard '{dashboard_name}' does not exist."}
        ),
        404,
    )


@dashboardsBP.errorhandler(InvalidFileNameError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The dashboard name provided is invalid. Try a name without special characters.",
            }
        ),
        400,
    )


@dashboardsBP.errorhandler(ScheduleNotFoundError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline for data source '{ex.namespace}' has not run yet. Please set up your connection.",
            }
        ),
        404,
    )


@dashboardsBP.route("/all", methods=["GET"])
def get_dashboards():
    response_data = dashboards_service().get_dashboards()
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/<dashboard_id>", methods=["GET"])
def get_dashboard(dashboard_id):
    response_data = dashboards_service().get_dashboard(dashboard_id)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/save", methods=["POST"])
@block_if_readonly
def save_dashboard():
    """
    Endpoint for saving a dashboard
    """
    post_data = request.get_json()
    response_data = dashboards_service().save_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/delete", methods=["DELETE"])
@block_if_readonly
def delete_dashboard():
    """
    Endpoint for deleting a dashboard
    """
    post_data = request.get_json()
    response_data = dashboards_service().delete_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/update", methods=["POST"])
@block_if_readonly
def update_dashboard():
    """
    Endpoint for updating a dashboard
    """
    post_data = request.get_json()
    response_data = dashboards_service().update_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/report/add", methods=["POST"])
@block_if_readonly
def add_report_to_dashboard():
    post_data = request.get_json()
    response_data = dashboards_service().add_report_to_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/report/remove", methods=["POST"])
@block_if_readonly
def remove_report_from_dashboard():
    post_data = request.get_json()
    response_data = dashboards_service().remove_report_from_dashboard(post_data)
    return jsonify(response_data)


@dashboardsBP.route("/dashboard/reports", methods=["POST"])
def get_dashboard_reports_with_query_results():
    dashboards_helper = DashboardsHelper()
    post_data = request.get_json()
    response_data = dashboards_helper.get_dashboard_reports_with_query_results(
        post_data
    )
    return jsonify(response_data)



import logging
import sqlparse

import networkx as nx

from copy import deepcopy
from collections import namedtuple
from enum import Enum
from functools import singledispatch
from networkx.readwrite import json_graph
from sqlparse.sql import TokenList, Comparison
from typing import Dict

from .base import MeltanoTable, MeltanoDesign
from .analysis_helper import AnalysisHelper
from meltano.core.behavior.visitor import visit_with


class InvalidIdentifier(Exception):
    pass


Identifier = namedtuple("Identifier", ("schema", "table", "field", "alias"))

DimensionGroup = namedtuple("DimensionGroup", ("dimensions", "group"))

ComparisonFields = namedtuple("ComparisonFields", ("left", "right"))


# Visitor pattern on the SQL AST
@singledispatch
def visit(node, executor, depth=0):
    logging.debug(f"Visiting node {node}")


@visit.register(TokenList)
def _(node: TokenList, executor, depth=0):
    logging.debug(f"Visiting node {node}")
    for t in node.tokens:
        executor.visit(t, depth + 1)


@visit.register(Comparison)
def _(node: Comparison, executor, depth=0):
    logging.debug(f"Visiting node {node}")
    executor.comparison(node, depth)


@visit_with(visit)
class PypikaJoinExecutor:
    def __init__(self, design, join):
        self.design = MeltanoDesign(design)
        self.join = join
        self.result = None
        self.comparison_fields = None

    def set_comparison_fields(self, left, right):
        self.comparison_fields = ComparisonFields(left, right)

    def comparison(self, node, depth):
        table = self.join["related_table"]

        left = self.parse_identifier(node.left)
        right = self.parse_identifier(node.right)

        left_table = self.design.find_table(left.table)
        right_table = self.design.find_table(right.table)

        left_field = getattr(
            AnalysisHelper.db_table(
                left_table.sql_table_name, alias=left_table.find_source_name()
            ),
            left.field,
        )

        right_field = getattr(
            AnalysisHelper.db_table(
                right_table.sql_table_name, alias=right_table.find_source_name()
            ),
            right.field,
        )

        self.set_comparison_fields(left, right)

        # TODO: do a stack based approach to handle complex cases
        self.result = left_field == right_field

    @classmethod
    def parse_identifier(cls, identifier) -> Identifier:
        token_values = map(lambda t: t.value, identifier.tokens)

        # name, field
        if len(identifier.tokens) == 3:
            table, _, field = token_values
            return Identifier(None, table, field, None)

        # schema, name, field
        if len(identifier.tokens) == 5:
            schema, _, table, _, field = token_values
            return Identifier(schema, table, field, None)

        # schema, name, field, alias
        if len(identifier.tokens) == 7:
            schema, _, table, _, field, _, alias = token_values
            return Identifier(schema, table, field, alias)

        raise InvalidIdentifier(str(identifier))


class JoinNotFound(Exception):
    pass


class JoinType(Enum):
    left_join = "left"
    inner_join = "inner"
    full_outer_join = "full_outer"
    cross_join = "cross"


class DesignHelper:
    def __init__(self, design):
        self.design = design

    @property
    def name(self):
        return self.design["name"]

    @property
    def joins(self):
        return deepcopy(self.design["joins"])

    @property
    def tables(self):
        tables = [self.design["related_table"]]
        tables += [j["related_table"] for j in self.design["joins"]]
        return list(map(MeltanoTable, tables))

    @property
    def base_table_name(self):
        return self.design["related_table"]["name"]

    def joins_for_table(self, table_name: str):
        # get the graph for networkx
        G = json_graph.node_link_graph(self.design["graph"])
        return nx.shortest_path(G, source=self.base_table_name, target=table_name)

    def join_for(self, join_selection: Dict):
        join = self.get_join(self, join_selection["name"])
        table = join["related_table"]

        db_table = AnalysisHelper.db_table(table["sql_table_name"], alias=join["name"])
        selected = {"columns": [], "aggregates": [], "timeframes": []}

        try:
            selected["columns"] = AnalysisHelper.columns_from_names(
                join_selection["columns"], table
            )
        except KeyError:
            pass

        try:
            selected["aggregates"] = AnalysisHelper.aggregates_from_names(
                join_selection["aggregates"], table
            )
        except KeyError:
            pass

        try:
            selected["timeframes"] = [
                self.timeframe_periods_for(table, timeframe_selection)
                for timeframe_selection in join_selection["timeframes"]
            ]
        except KeyError:
            pass

        join_executor = PypikaJoinExecutor(self.design, join)
        join_executor.visit(sqlparse.parse(join["sql_on"])[0])

        return {
            "table": table,
            "db_table": db_table,
            "on": join_executor.result,
            "join": join,
            "name": join_selection["name"],
            **selected,
        }

    def timeframe_periods_for(self, table, timeframe_selection):
        timeframe_name = timeframe_selection["name"]
        period_names = [p["label"] for p in timeframe_selection["periods"]]

        db_table = AnalysisHelper.db_table(
            table["sql_table_name"], alias=timeframe_name
        )
        timeframe, periods = AnalysisHelper.timeframe_periods_from_names(
            timeframe_name, period_names, table
        )

        return {
            "table": table,
            "db_table": db_table,
            "timeframe": timeframe,
            "periods": periods,
            "period_labels": period_names,
        }

    @classmethod
    def get_join(cls, design, name):
        try:
            return next(join for join in design.joins if join["name"] == name)
        except StopIteration:
            raise JoinNotFound(
                f"No join named '{name}' was found in design '{design.name}'."
            )

    def __getitem__(self, idx):
        return self.design[idx]



import os
import json
import yaml
import logging
from pathlib import Path

from .project import Project
from .config_service import ConfigService
from .plugin.settings_service import PluginSettingsService
from .plugin_discovery_service import PluginDiscoveryService
from .plugin import PluginDefinition, PluginType, ProjectPlugin
from .db import project_engine


class TransformAddService:
    def __init__(self, project: Project):
        self.project = project

        self.config_service = ConfigService(project)
        dbt_plugin = self.config_service.find_plugin(
            plugin_name="dbt", plugin_type=PluginType.TRANSFORMERS
        )

        self.discovery_service = PluginDiscoveryService(
            self.project, config_service=self.config_service
        )
        settings_service = PluginSettingsService(
            project,
            dbt_plugin,
            config_service=self.config_service,
            plugin_discovery_service=self.discovery_service,
        )
        dbt_project_dir = settings_service.get("project_dir")
        dbt_project_path = Path(dbt_project_dir)

        self.packages_file = dbt_project_path.joinpath("packages.yml")
        self.dbt_project_file = dbt_project_path.joinpath("dbt_project.yml")

    def add_to_packages(self, plugin: ProjectPlugin):
        if not os.path.exists(self.packages_file):
            with open(self.packages_file, "w"):
                pass

        package_yaml = yaml.safe_load(self.packages_file.open()) or {"packages": []}

        for package in package_yaml["packages"]:
            if package.get("git", "") == plugin.pip_url:
                return

        package_yaml["packages"].append({"git": plugin.pip_url})

        with open(self.packages_file, "w") as f:
            f.write(yaml.dump(package_yaml, default_flow_style=False, sort_keys=False))

    def update_dbt_project(self, plugin: ProjectPlugin):
        settings_service = PluginSettingsService(
            self.project,
            plugin,
            config_service=self.config_service,
            plugin_discovery_service=self.discovery_service,
        )

        package_name = settings_service.get("_package_name")
        vars = settings_service.get("_vars")

        model_def = {}
        if vars:
            model_def["vars"] = vars

        dbt_project_yaml = yaml.safe_load(self.dbt_project_file.open())
        dbt_project_yaml["models"][package_name] = model_def

        with open(self.dbt_project_file, "w") as f:
            f.write(
                yaml.dump(dbt_project_yaml, default_flow_style=False, sort_keys=False)
            )



import yaml
import copy


from typing import Iterable, List
from meltano.core.schedule import Schedule
from meltano.core.plugin import ProjectPlugin, PluginType
from meltano.core.plugin.factory import plugin_factory
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq


VERSION = 1


class MeltanoFile(Canonical):
    def __init__(
        self, version: int = VERSION, plugins={}, schedules: list = [], **extras
    ):
        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            version=version,
            extras=extras,
            plugins=self.load_plugins(plugins),
            schedules=self.load_schedules(schedules),
        )

    def load_plugins(self, plugins) -> Canonical:
        """Parse the meltano.yml file and return it as `ProjectPlugin` instances."""
        plugin_type_plugins = Canonical()

        for plugin_type in PluginType:
            plugin_type_plugins[plugin_type] = []

        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        for plugin_type, raw_plugins in plugins.items():
            for raw_plugin in raw_plugins:
                plugin = plugin_factory(plugin_type, raw_plugin)
                plugin_type_plugins[plugin.type].append(plugin)

        return plugin_type_plugins

    def load_schedules(self, schedules) -> List[Schedule]:
        return list(map(Schedule.parse, schedules))



import logging
import subprocess
import asyncio
import os
import copy
from typing import Optional
from contextlib import contextmanager

from .project import Project
from .plugin import PluginRef, ProjectPlugin
from .plugin.error import PluginMissingError, PluginExecutionError
from .plugin.config_service import PluginConfigService
from .plugin.settings_service import PluginSettingsService
from .plugin_discovery_service import PluginDiscoveryService
from .venv_service import VenvService, VirtualEnv
from .error import Error, SubprocessError
from .logging.utils import OUTPUT_BUFFER_SIZE


def invoker_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    cls = PluginInvoker

    if hasattr(plugin.__class__, "__invoker_cls__"):
        cls = plugin.__class__.__invoker_cls__

    invoker = cls(project, plugin, *args, **kwargs)

    return invoker


class InvokerError(Error):
    pass


class ExecutableNotFoundError(InvokerError):
    """Occurs when the executable could not be found"""

    def __init__(self, plugin: PluginRef, executable):
        super().__init__(
            f"Executable '{executable}' could not be found. "
            f"{plugin.type.descriptor.capitalize()} '{plugin.name}' may not have been installed yet using `meltano install {plugin.type.singular} {plugin.name}`, or the executable name may be incorrect."
        )


class InvokerNotPreparedError(InvokerError):
    """Occurs when `invoke` is called before `prepare`"""

    pass


class PluginInvoker:
    """This class handles the invocation of a `ProjectPlugin` instance."""

    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        context: Optional[object] = None,
        run_dir=None,
        config_dir=None,
        venv_service: VenvService = None,
        plugin_config_service: PluginConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
        plugin_discovery_service: PluginDiscoveryService = None,
    ):
        self.project = project
        self.plugin = plugin
        self.context = context

        self.venv_service = venv_service or VenvService(project)
        self.config_service = plugin_config_service or PluginConfigService(
            plugin,
            config_dir or self.project.plugin_dir(plugin),
            run_dir or self.project.run_dir(plugin.name),
        )
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.settings_service = plugin_settings_service or PluginSettingsService(
            project,
            plugin,
            config_service=self.discovery_service.config_service,
            plugin_discovery_service=self.discovery_service,
        )

        self.plugin_def = self.discovery_service.get_definition(plugin)

        self._prepared = False
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

    @property
    def capabilities(self):
        # we want to make sure the capabilites are immutable from the `PluginInvoker` interface
        return frozenset(self.plugin_def.capabilities)

    @property
    def files(self):
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}

        return {
            _key: self.config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    def prepare(self, session):
        self.plugin_config = self.settings_service.as_dict(
            extras=False, session=session
        )
        self.plugin_config_processed = self.settings_service.as_dict(
            extras=False, process=True, session=session
        )
        self.plugin_config_extras = self.settings_service.as_dict(
            extras=True, session=session
        )
        self.plugin_config_env = self.settings_service.as_env(session=session)

        with self.plugin.trigger_hooks("configure", self, session):
            self.config_service.configure()
            self._prepared = True

    def cleanup(self):
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

        with self.plugin.trigger_hooks("cleanup", self):
            self._prepared = False

    @contextmanager
    def prepared(self, session):
        try:
            self.prepare(session)
            yield
        finally:
            self.cleanup()

    @property
    def executable(self):
        return self.plugin_def.executable or self.plugin.name

    def exec_path(self):
        return self.venv_service.exec_path(
            self.executable, name=self.plugin.name, namespace=self.plugin.type
        )

    def exec_args(self, *args):
        plugin_args = self.plugin.exec_args(self)

        return [str(arg) for arg in (self.exec_path(), *plugin_args, *args)]

    def env(self):
        env = {
            **self.project.dotenv_env,
            **self.settings_service.env,
            **self.plugin_config_env,
        }

        # Ensure Meltano venv is not inherited
        venv = VirtualEnv(self.project.venvs_dir(self.plugin.type, self.plugin.name))
        env["VIRTUAL_ENV"] = str(venv.root)
        env["PATH"] = os.pathsep.join([str(venv.bin_dir), env["PATH"]])
        env.pop("PYTHONPATH", None)

        return env

    def Popen_options(self):
        return {}

    @contextmanager
    def _invoke(self, *args, require_preparation=True, env={}, **Popen):
        if require_preparation and not self._prepared:
            raise InvokerNotPreparedError()

        with self.plugin.trigger_hooks("invoke", self, args):
            popen_args = self.exec_args(*args)
            popen_options = {**self.Popen_options(), **Popen}
            popen_env = {**self.env(), **env}
            logging.debug(f"Invoking: {popen_args}")
            logging.debug(f"Env: {popen_env}")

            try:
                yield (popen_args, popen_options, popen_env)
            except FileNotFoundError as err:
                raise ExecutableNotFoundError(self.plugin, self.executable) from err

    def invoke(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return subprocess.Popen(popen_args, **popen_options, env=popen_env)

    async def invoke_async(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return await asyncio.create_subprocess_exec(
                *popen_args, limit=OUTPUT_BUFFER_SIZE, **popen_options, env=popen_env
            )

    def dump(self, file_id):
        try:
            with self._invoke():
                return self.files[file_id].read_text()
        except ExecutableNotFoundError as err:
            # Unwrap FileNotFoundError
            raise err.__cause__



from enum import Enum
from flask_mail import Mail, Message
from smtpapi import SMTPAPIHeader

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.api.models.subscription import Subscription, SubscriptionEventType


mail = Mail()


# From `https://app.sendgrid.com` Meltano's account
class UnsubscribeGroup(int, Enum):
    ELT_NOTIFICATIONS = 12751


class MailService:
    EventTypeUnsubscribeGroup = {
        SubscriptionEventType.PIPELINE_MANUAL_RUN: UnsubscribeGroup.ELT_NOTIFICATIONS
    }

    def __init__(self, project: Project):
        self.project = project
        self.project_id = ProjectSettingsService(self.project).get("project_id")

    def get_unsubscribe_group(self, subscription) -> UnsubscribeGroup:
        try:
            event_type = SubscriptionEventType(subscription.event_type)
            return self.EventTypeUnsubscribeGroup[event_type]
        except:
            return None

    def create_message(self, subscription, **kwargs) -> Message:
        """
        Create the basic Message with the proper headers.

        Uses the SendGrid SMTPAPI.
        """

        headers = SMTPAPIHeader()
        headers.set_unique_args(
            {
                "project_id": str(self.project_id),
                "subscription_id": str(subscription.id),
            }
        )

        unsubscribe_group = self.get_unsubscribe_group(subscription)
        if unsubscribe_group:
            headers.set_asm_group_id(int(unsubscribe_group))

        return Message(
            recipients=[subscription.recipient],
            extra_headers={"X-SMTPAPI": headers.json_string()},
            **kwargs
        )



import os
import yaml
import click
import logging
from urllib.parse import urlparse

from meltano.core.project_init_service import (
    ProjectInitService,
    ProjectInitServiceError,
)
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.error import SubprocessError
from . import cli
from .utils import CliError
from .params import database_uri_option

EXTRACTORS = "extractors"
LOADERS = "loaders"
ALL = "all"

logger = logging.getLogger(__name__)


@cli.command()
@click.pass_context
@click.argument("project_name")
@click.option(
    "--no_usage_stats", help="Do not send anonymous usage stats.", is_flag=True
)
@database_uri_option
def init(ctx, project_name, no_usage_stats):
    """
    Creates a new Meltano project
    """
    if ctx.obj["project"]:
        logging.warning(f"Found meltano project at: {ctx.obj['project'].root}")
        raise CliError("`meltano init` cannot run inside a Meltano project.")

    if no_usage_stats:
        ProjectSettingsService.config_override["send_anonymous_usage_stats"] = False

    init_service = ProjectInitService(project_name)
    try:
        project = init_service.init()
        init_service.echo_instructions()

        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_init(project_name=project_name)
    except ProjectInitServiceError as e:
        raise CliError(str(e)) from e
    except SubprocessError as err:
        logger.error(err.stderr)
        raise CliError(str(err)) from err



import click
import sys
import logging
from . import cli
from .params import project
from .utils import CliError

from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import invoker_factory, InvokerError
from meltano.core.config_service import ConfigService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine
from meltano.core.error import SubprocessError


logger = logging.getLogger(__name__)


@cli.command(
    context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False)
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.argument("plugin_name")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
@project(migrate=True)
def invoke(project, plugin_type, plugin_name, plugin_args):
    plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

    _, Session = project_engine(project)
    session = Session()
    try:
        config_service = ConfigService(project)
        plugin = config_service.find_plugin(
            plugin_name, plugin_type=plugin_type, invokable=True
        )
        invoker = invoker_factory(project, plugin)
        with invoker.prepared(session):
            handle = invoker.invoke(*plugin_args)
            exit_code = handle.wait()

        tracker = GoogleAnalyticsTracker(project)
        tracker.track_meltano_invoke(
            plugin_name=plugin_name, plugin_args=" ".join(plugin_args)
        )

        sys.exit(exit_code)
    except InvokerError as err:
        raise CliError(str(err)) from err
    except SubprocessError as err:
        logger.error(err.stderr)
        raise CliError(str(err)) from err
    finally:
        session.close()



import os
import io
import yaml
import requests
import logging
import shutil
import re
from copy import deepcopy
from typing import Dict, Iterable, Optional

import meltano.core.bundle as bundle
from .project_settings_service import ProjectSettingsService
from .setting_definition import SettingDefinition
from .behavior.versioned import Versioned, IncompatibleVersionError
from .behavior.canonical import Canonical
from .config_service import ConfigService
from .plugin import PluginDefinition, ProjectPlugin, PluginType, PluginRef, Variant
from .plugin.factory import plugin_factory


class PluginNotFoundError(Exception):
    pass


class DiscoveryInvalidError(Exception):
    """Occurs when the discovery.yml fails to be parsed."""

    pass


class DiscoveryUnavailableError(Exception):
    """Occurs when the discovery.yml cannot be found or downloaded."""

    pass


# Increment this version number whenever the schema of discovery.yml is changed.
# See https://www.meltano.com/docs/contributor-guide.html#discovery-yml-version for more information.
VERSION = 16


class DiscoveryFile(Canonical):
    def __init__(self, version=1, **plugins):
        super().__init__(version=int(version))

        for plugin_type in PluginType:
            self[plugin_type] = []

        for plugin_type, raw_plugins in plugins.items():
            for raw_plugin in raw_plugins:
                plugin_def = PluginDefinition(
                    plugin_type,
                    raw_plugin.pop("name"),
                    raw_plugin.pop("namespace"),
                    **raw_plugin,
                )
                self[plugin_type].append(plugin_def)

    @classmethod
    def version(cls, attrs):
        return int(attrs.get("version", 1))


class PluginDiscoveryService(Versioned):
    __version__ = VERSION

    def __init__(
        self,
        project,
        config_service: ConfigService = None,
        discovery: Optional[Dict] = None,
    ):
        self.project = project
        self.config_service = config_service or ConfigService(project)

        self._discovery_version = None
        self._discovery = None
        if discovery:
            self._discovery_version = DiscoveryFile.version(discovery)
            self._discovery = DiscoveryFile.parse(discovery)

    @property
    def file_version(self):
        return self._discovery_version

    @property
    def discovery_url(self):
        discovery_url = ProjectSettingsService(self.project).get("discovery_url")

        if not discovery_url or not re.match(r"^https?://", discovery_url):
            return None

        return discovery_url

    @property
    def discovery(self):
        """
        Return first compatible discovery manifest from these locations:

        - project local `discovery.yml`
        - `discovery_url` project setting
        - .meltano/cache/discovery.yml
        - meltano.core.bundle
        """
        if self._discovery:
            return self._discovery

        loaders = [
            (
                self.load_local_discovery,
                "your project's local `discovery.yml` manifest",
            ),
            (
                self.load_remote_discovery,
                f"the `discovery.yml` manifest received from {self.discovery_url}",
            ),
            (self.load_cached_discovery, "the cached `discovery.yml` manifest"),
            (self.load_bundled_discovery, "the bundled `discovery.yml` manifest"),
        ]

        errored = False
        for loader, description in loaders:
            if errored:
                logging.warning(f"Falling back on {description}...")

            try:
                loader()
            except IncompatibleVersionError as err:
                errored = True

                logging.warning(
                    f"{description.capitalize()} has version {err.file_version}, while this version of Meltano requires version {err.version}."
                )
                if err.file_version > err.version:
                    logging.warning(
                        "Please install the latest compatible version of Meltano using `meltano upgrade`."
                    )
            except DiscoveryInvalidError as err:
                errored = True
                logging.error(f"{description.capitalize()} could not be parsed.")
                logging.debug(str(err))

            if self._discovery:
                return self._discovery

        raise DiscoveryInvalidError("No valid `discovery.yml` manifest could be found")

    def load_local_discovery(self):
        try:
            with self.project.root_dir("discovery.yml").open() as local_discovery:
                return self.load_discovery(local_discovery)
        except FileNotFoundError:
            pass

    def load_remote_discovery(self):
        if not self.discovery_url:
            return

        try:
            response = requests.get(self.discovery_url)
            response.raise_for_status()

            remote_discovery = io.StringIO(response.text)
            discovery = self.load_discovery(remote_discovery, cache=True)

            return discovery
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as err:
            logging.debug("Remote `discovery.yml` manifest could not be downloaded.")
            logging.debug(str(err))
            pass

    def load_cached_discovery(self):
        try:
            with self.cached_discovery_file.open() as cached_discovery:
                return self.load_discovery(cached_discovery)
        except FileNotFoundError:
            pass

    def load_bundled_discovery(self):
        with bundle.find("discovery.yml").open() as bundled_discovery:
            discovery = self.load_discovery(bundled_discovery, cache=True)

        return discovery

    def load_discovery(self, discovery_file, cache=False):
        try:
            discovery_yaml = yaml.safe_load(discovery_file)

            self._discovery_version = DiscoveryFile.version(discovery_yaml)
            self.ensure_compatible()

            self._discovery = DiscoveryFile.parse(discovery_yaml)

            if cache:
                self.cache_discovery()

            return self._discovery
        except IncompatibleVersionError:
            raise
        except (yaml.YAMLError, Exception) as err:
            raise DiscoveryInvalidError(str(err))

    def cache_discovery(self):
        with self.cached_discovery_file.open("w") as cached_discovery:
            yaml.dump(
                self._discovery,
                cached_discovery,
                default_flow_style=False,
                sort_keys=False,
            )

    @property
    def cached_discovery_file(self):
        return self.project.meltano_dir("cache", "discovery.yml")

    def get_discovery_plugins_of_type(self, plugin_type):
        return self.discovery[plugin_type]

    def get_custom_plugins_of_type(self, plugin_type):
        return [
            project_plugin.custom_definition
            for project_plugin in self.config_service.get_plugins_of_type(plugin_type)
            if project_plugin.is_custom()
        ]

    def get_plugins_of_type(self, plugin_type):
        return self.get_custom_plugins_of_type(
            plugin_type
        ) + self.get_discovery_plugins_of_type(plugin_type)

    def plugins_by_type(self):
        return {
            plugin_type: self.get_plugins_of_type(plugin_type)
            for plugin_type in PluginType
        }

    def plugins(self) -> Iterable[PluginDefinition]:
        yield from (
            plugin
            for plugin_type, plugins in self.plugins_by_type().items()
            for plugin in plugins
        )

    def find_definition(
        self, plugin_type: PluginType, plugin_name: str, variant=None
    ) -> PluginDefinition:
        name, _ = PluginRef.parse_name(plugin_name)
        try:
            plugin = next(
                plugin
                for plugin in self.get_plugins_of_type(plugin_type)
                if plugin.name == name
            )
            plugin.use_variant(variant)
            return plugin
        except StopIteration as stop:
            raise PluginNotFoundError(name) from stop

    def find_definition_by_namespace(
        self, plugin_type: PluginType, namespace: str
    ) -> PluginDefinition:
        try:
            return next(
                plugin
                for plugin in self.get_plugins_of_type(plugin_type)
                if plugin.namespace == namespace
            )
        except StopIteration as stop:
            raise PluginNotFoundError(namespace) from stop

    def get_definition(self, project_plugin: ProjectPlugin) -> PluginDefinition:
        if project_plugin.is_custom():
            plugin = project_plugin.custom_definition
        else:
            try:
                plugin = next(
                    plugin for plugin in self.plugins() if plugin == project_plugin
                )
            except StopIteration as stop:
                raise PluginNotFoundError(project_plugin.name) from stop

        plugin.use_variant(project_plugin.variant or Variant.ORIGINAL_NAME)

        return plugin



import logging
import requests
import threading
import time
import webbrowser
import click

from meltano.core.project_settings_service import ProjectSettingsService


class UIAvailableWorker(threading.Thread):
    def __init__(self, project, open_browser=False):
        super().__init__()

        self.project = project
        self.open_browser = open_browser

        self.settings_service = ProjectSettingsService(self.project)

        self._terminate = False

    def run(self):
        bind_port = self.settings_service.get("ui.bind_port")
        url = f"http://localhost:{bind_port}"

        server_name = self.settings_service.get("ui.server_name")
        headers = {"Host": server_name}

        while not self._terminate:
            try:
                response = requests.get(url, headers=headers)
                if response.status_code == 200:
                    click.secho(f"Meltano UI is now available at {url}", fg="green")
                    if self.open_browser:
                        webbrowser.open(url)
                    self._terminate = True
            except:
                pass

            time.sleep(2)

    def stop(self):
        self._terminate = True



import os
import json
import yaml
import logging
from typing import List

from .project import Project
from .plugin import PluginType, PluginDefinition, ProjectPlugin, PluginRef
from .plugin_discovery_service import PluginDiscoveryService
from .plugin.factory import plugin_factory
from .config_service import ConfigService, PluginAlreadyAddedException


class PluginNotSupportedException(Exception):
    pass


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    def __init__(
        self,
        project: Project,
        plugin_discovery_service: PluginDiscoveryService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.config_service = config_service or ConfigService(project)

    def add(self, *args, **kwargs) -> ProjectPlugin:
        plugin_def = self.discovery_service.find_definition(*args, **kwargs)
        return self.add_definition(plugin_def)

    def add_definition(self, plugin_def: PluginDefinition, **kwargs) -> ProjectPlugin:
        plugin = plugin_def.in_project(**kwargs)
        return self.config_service.add_to_file(plugin)

    def add_related(
        self,
        target_plugin: ProjectPlugin,
        plugin_types: List[PluginType] = list(PluginType),
    ):
        try:
            plugin_types.remove(target_plugin.type)
        except ValueError:
            pass

        related_plugin_refs = []

        runner_ref = target_plugin.runner
        if runner_ref:
            related_plugin_refs.append(runner_ref)

        plugin_def = self.discovery_service.get_definition(target_plugin)
        related_plugin_refs.extend(
            related_plugin_def
            for plugin_type in plugin_types
            for related_plugin_def in self.discovery_service.get_plugins_of_type(
                plugin_type
            )
            if related_plugin_def.namespace == plugin_def.namespace
        )

        added_plugins = []
        for plugin_ref in related_plugin_refs:
            try:
                plugin = self.add(plugin_ref.type, plugin_ref.name)
            except PluginAlreadyAddedException as err:
                continue

            added_plugins.append(plugin)

        added_plugins_with_related = []
        for plugin in added_plugins:
            added_plugins_with_related.extend(
                [plugin, *self.add_related(plugin, plugin_types=plugin_types)]
            )

        return added_plugins_with_related



import urllib.parse
from datetime import timedelta
from functools import wraps
from flask import current_app, request, redirect, jsonify, make_response
from flask_login import current_user, user_logged_in
from flask_security import Security, login_required
from flask_security.utils import login_user
from flask_principal import identity_loaded, Identity

from meltano.core.project_settings_service import ProjectSettingsService
from .identity import users, FreeUser, create_dev_user
from .forms import MeltanoLoginForm, MeltanoRegisterFrom, MeltanoConfirmRegisterForm
from .auth import (
    unauthorized_callback,
    _user_logged_in_hook,
    _identity_loaded_hook,
    block_if_api_auth_required,
)


# normally one would setup the extension accordingly, but it
# seems Security.init_app() overwrites all the configuration
security = Security()


def setup_security(app, project):
    options = {
        "login_form": MeltanoLoginForm,
        "register_form": MeltanoRegisterFrom,
        "confirm_register_form": MeltanoConfirmRegisterForm,
    }

    settings_service = ProjectSettingsService(project)
    if not settings_service.get("ui.authentication"):
        # the FreeUser is free to do everything and has all
        # roles and permissions automatically.
        options["anonymous_user"] = FreeUser
    else:
        # Use Flask's built-in AnonymousUser which is not deemed to be authenticated
        # and has no roles
        pass

    security.init_app(app, users, **options)
    security.unauthorized_handler(unauthorized_callback)
    user_logged_in.connect_via(app)(_user_logged_in_hook)
    identity_loaded.connect_via(app)(_identity_loaded_hook)



import os
import click
import logging
import psutil
import importlib
import meltano
import subprocess
import signal
import shutil
from typing import Optional

from meltano.core.project import Project
from meltano.core.migration_service import MigrationService, MigrationError
from meltano.core.config_service import ConfigService
from meltano.cli.utils import install_plugins, PluginInstallReason
from meltano.core.compiler.project_compiler import ProjectCompiler
import meltano.core.bundle as bundle


class UpgradeError(Exception):
    """Occurs when the Meltano upgrade fails"""

    pass


class AutomaticPackageUpgradeError(Exception):
    def __init__(self, reason, instructions):
        self.reason = reason
        self.instructions = instructions


class UpgradeService:
    def __init__(self, engine, project: Project):
        self.project = project
        self.engine = engine

    def reload_ui(self):
        click.secho("Reloading UI...", fg="blue")

        pid_file_path = self.project.run_dir("gunicorn.pid")
        try:
            with pid_file_path.open("r") as pid_file:
                pid = int(pid_file.read())

                process = psutil.Process(pid)
                process.send_signal(signal.SIGHUP)
        except FileNotFoundError:
            click.echo("UI is not running")
        except Exception as ex:
            logging.error(f"Cannot restart from `{pid_file_path}`: {ex}")

    def _upgrade_package(self, pip_url: Optional[str] = None, force=False):
        meltano_file_path = "/src/meltano/__init__.py"
        editable = meltano.__file__.endswith(meltano_file_path)
        if editable and not force:
            meltano_dir = meltano.__file__[0 : -len(meltano_file_path)]
            raise AutomaticPackageUpgradeError(
                reason="it is installed from source",
                instructions=f"navigate to `{meltano_dir}` and run `git pull`",
            )

        in_docker = os.path.exists("/.dockerenv")
        if in_docker:
            raise AutomaticPackageUpgradeError(
                reason="it is installed inside Docker",
                instructions="pull the latest Docker image using `docker pull meltano/meltano` and recreate any containers you may have created",
            )

        pip_url = pip_url or "meltano"
        run = subprocess.run(
            ["pip", "install", "--upgrade", pip_url],
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        if run.returncode != 0:
            raise UpgradeError(f"Failed to upgrade `meltano`.", run)

        return True

    def upgrade_package(self, *args, **kwargs):
        click.secho("Upgrading `meltano` package...", fg="blue")

        try:
            self._upgrade_package(*args, **kwargs)
            click.echo("The `meltano` package has been upgraded.")

            click.echo()
            self.reload_ui()

            return True
        except AutomaticPackageUpgradeError as err:
            click.echo(
                f"{click.style('The `meltano` package could not be upgraded automatically', fg='red')} because {err.reason}."
            )
            click.echo(f"To upgrade manually, {err.instructions}.")
            return False

    def update_files(self):
        """
        Update the files managed by Meltano inside the current project.
        """
        click.secho("Updating files managed by plugins...", fg="blue")

        file_plugins = ConfigService(self.project).get_files()
        if not file_plugins:
            click.echo("Nothing to update")
            return

        install_plugins(self.project, file_plugins, reason=PluginInstallReason.UPGRADE)

    def migrate_database(self):
        click.secho("Applying migrations to system database...", fg="blue")

        try:
            migration_service = MigrationService(self.engine)
            migration_service.upgrade()
            migration_service.seed(self.project)
        except MigrationError as err:
            raise UpgradeError(str(err)) from err

    def compile_models(self):
        click.secho("Recompiling models...", fg="blue")

        ProjectCompiler(self.project).compile()

    def upgrade(self, skip_package=False, **kwargs):
        package_upgraded = False
        if not skip_package:
            package_upgraded = self.upgrade_package(**kwargs)

            if not package_upgraded:
                click.echo(
                    "Then, run `meltano upgrade --skip-package` to upgrade your project based on the latest version."
                )
                return

            click.echo()

        self.update_files()
        click.echo()
        self.migrate_database()
        click.echo()
        self.compile_models()

        click.echo()
        if package_upgraded:
            click.secho(
                "Meltano and your Meltano project have been upgraded!", fg="green"
            )
        else:
            click.secho("Your Meltano project has been upgraded!", fg="green")



from typing import Dict

from meltano.core.behavior.hookable import hook
from meltano.core.db import project_engine
from meltano.core.setting_definition import SettingDefinition
from . import SingerPlugin, PluginType


class SingerTarget(SingerPlugin):
    __plugin_type__ = PluginType.LOADERS

    @property
    def extra_settings(self):
        return [
            SettingDefinition(name="_dialect", value="$MELTANO_LOADER_NAMESPACE"),
            SettingDefinition(name="_target_schema", value="$MELTANO_LOAD_SCHEMA"),
        ]

    def exec_args(self, plugin_invoker):
        args = ["--config", plugin_invoker.files["config"]]

        return args

    @property
    def config_files(self):
        return {"config": "target.config.json"}

    @property
    def output_files(self):
        return {"state": "new_state.json"}



import json
from typing import List
from collections import OrderedDict

from .utils import truthy, flatten, nest_object, to_env_var
from .behavior.canonical import Canonical
from .behavior import NameEq
from .error import Error

VALUE_PROCESSORS = {"nest_object": nest_object, "upcase_string": lambda s: s.upper()}


class EnvVar:
    def __init__(self, definition):
        key = definition
        negated = False

        if definition.startswith("!"):
            key = definition[1:]
            negated = True

        self.key = key
        self.negated = negated

    @property
    def definition(self):
        prefix = "!" if self.negated else ""
        return f"{prefix}{self.key}"

    def get(self, env):
        if self.negated:
            return str(not truthy(env[self.key]))
        else:
            return env[self.key]


class SettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, name: str):
        super().__init__(f"Cannot find setting {name}")


class SettingDefinition(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        aliases: List[str] = [],
        env: str = None,
        env_aliases: List[str] = [],
        kind: str = None,
        value=None,
        label: str = None,
        documentation: str = None,
        description: str = None,
        tooltip: str = None,
        options: list = [],
        oauth: dict = {},
        placeholder: str = None,
        protected: bool = None,
        env_specific: bool = None,
        custom: bool = False,
        value_processor=None,
        **attrs,
    ):
        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            name=name,
            aliases=aliases,
            env=env,
            env_aliases=env_aliases,
            kind=kind,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            protected=protected,
            env_specific=env_specific,
            value_processor=value_processor,
            _custom=custom,
            **attrs,
        )

        self._verbatim.add("value")

    @classmethod
    def from_missing(cls, defs, config, **kwargs):
        flat_config = flatten(config, "dot")

        names = set(s.name for s in defs)

        # Create custom setting definitions for unknown keys
        return [
            SettingDefinition.from_key_value(k, v, **kwargs)
            for k, v in flat_config.items()
            if k not in names
        ]

    @classmethod
    def from_key_value(cls, key, value, custom=True, default=False):
        kind = None
        if isinstance(value, bool):
            kind = "boolean"
        elif isinstance(value, int):
            kind = "integer"
        elif isinstance(value, dict):
            kind = "object"
        elif isinstance(value, list):
            kind = "array"

        attrs = {
            "name": key,
            "kind": kind,
            "custom": custom,
            "value": value if default else None,
        }

        return cls(**attrs)

    @property
    def is_extra(self):
        return self.name.startswith("_")

    @property
    def is_redacted(self):
        return self.kind in ("password", "oauth")

    def env_vars(self, prefixes: [str]):
        env_keys = []

        if self.env:
            env_keys.append(self.env)

        env_keys.extend(to_env_var(prefix, self.name) for prefix in prefixes)

        env_keys.extend(alias for alias in self.env_aliases)

        # Drop duplicate keys
        env_keys = list(OrderedDict.fromkeys(env_keys))

        return [EnvVar(key) for key in env_keys]

    def cast_value(self, value):
        if isinstance(value, str):
            if self.kind == "boolean":
                return truthy(value)
            elif self.kind == "integer":
                return int(value)
            elif self.kind == "object":
                value = json.loads(value)
                if not isinstance(value, dict):
                    raise ValueError(f"JSON value '{value}' is not an object")
            elif self.kind == "array":
                value = json.loads(value)
                if not isinstance(value, list):
                    raise ValueError(f"JSON value '{value}' is not an array")

        processor = self.value_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def stringify_value(self, value):
        if isinstance(value, str):
            return value

        if self.kind == "string":
            return str(value)

        return json.dumps(value)



import logging
from typing import Iterable
from enum import Enum

from .compiler.project_compiler import ProjectCompiler
from .plugin_discovery_service import PluginDiscoveryService
from .project_add_service import ProjectAddService
from .config_service import ConfigService
from .venv_service import VenvService
from .utils import noop
from .plugin import PluginType, ProjectPlugin, PluginDefinition, PluginRef
from .project import Project
from .error import (
    PluginInstallError,
    PluginInstallWarning,
    SubprocessError,
    PluginNotInstallable,
)


class PluginInstallReason(str, Enum):
    ADD = "add"
    INSTALL = "install"
    UPGRADE = "upgrade"


def installer_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    cls = PipPluginInstaller

    if hasattr(plugin.__class__, "__installer_cls__"):
        cls = plugin.__class__.__installer_cls__

    return cls(project, plugin, *args, **kwargs)


class PluginInstallService:
    def __init__(self, project: Project, config_service: ConfigService = None):
        self.project = project
        self.config_service = config_service or ConfigService(project)

    def install_all_plugins(self, reason=PluginInstallReason.INSTALL, status_cb=noop):
        return self.install_plugins(
            self.config_service.plugins(), reason=reason, status_cb=status_cb
        )

    def install_plugins(
        self,
        plugins: Iterable[ProjectPlugin],
        reason=PluginInstallReason.INSTALL,
        status_cb=noop,
    ):
        errors = []
        installed = []
        has_model = False

        for plugin in plugins:
            status = {"plugin": plugin, "status": "running"}
            status_cb(status, reason)

            try:
                self.install_plugin(plugin, compile_models=False, reason=reason)

                if plugin.type is PluginType.MODELS:
                    has_model = True

                status["status"] = "success"
                installed.append(status)
            except PluginInstallError as err:
                status["status"] = "error"
                status["message"] = str(err)
                status["details"] = err.stderr
                errors.append(status)
            except PluginInstallWarning as warn:
                status["status"] = "warning"
                status["message"] = str(warn)
                errors.append(status)
            except PluginNotInstallable as info:
                # let's totally ignore these plugins
                pass

            status_cb(status, reason)

        if has_model:
            self.compile_models()

        return {"errors": errors, "installed": installed}

    def install_plugin(
        self,
        plugin: ProjectPlugin,
        reason=PluginInstallReason.INSTALL,
        compile_models=True,
    ):
        if not plugin.is_installable():
            raise PluginNotInstallable()

        try:
            with plugin.trigger_hooks("install", self.project, reason):
                run = installer_factory(self.project, plugin).install(reason)

                if compile_models and plugin.type is PluginType.MODELS:
                    self.compile_models()

                return run
        except PluginInstallError:
            raise
        except SubprocessError as err:
            raise PluginInstallError(
                f"{plugin.type.descriptor} '{plugin.name}' could not be installed: {err}".capitalize(),
                err.process,
            ) from err

    def compile_models(self):
        compiler = ProjectCompiler(self.project)
        try:
            compiler.compile()
        except Exception:
            pass


class PipPluginInstaller:
    def __init__(
        self, project, plugin: ProjectPlugin, venv_service: VenvService = None
    ):
        self.plugin = plugin
        self.venv_service = venv_service or VenvService(project)

    def install(self, reason):
        self.venv_service.clean(namespace=self.plugin.type, name=self.plugin.name)
        self.venv_service.create(namespace=self.plugin.type, name=self.plugin.name)
        return self.venv_service.install(
            namespace=self.plugin.type,
            name=self.plugin.name,
            pip_url=self.plugin.pip_url,
        )



import json
import os
import sqlalchemy

from datetime import date, datetime
from decimal import Decimal
from flask import jsonify, request

from .settings_helper import SettingsHelper
from .sql_helper import SqlHelper, ConnectionNotFound, UnsupportedConnectionDialect
from meltano.api.api_blueprint import APIBlueprint
from meltano.core.project import Project
from meltano.core.schedule_service import ScheduleService, ScheduleNotFoundError
from meltano.core.utils import find_named, NotFound
from meltano.core.sql.filter import FilterOptions
from meltano.core.sql.base import ParseError, EmptyQuery


sqlBP = APIBlueprint("sql", __name__)


@sqlBP.errorhandler(ConnectionNotFound)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"Missing connection details to '{ex.connection_name}'. Create a connection to '{ex.connection_name}' in the settings.",
            }
        ),
        500,
    )


@sqlBP.errorhandler(UnsupportedConnectionDialect)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The dialect '{ex.connection_dialect}' is not yet supported by Meltano Analyze.",
            }
        ),
        500,
    )


@sqlBP.errorhandler(sqlalchemy.exc.DBAPIError)
def _handle(ex):
    return (
        jsonify(
            {"error": True, "code": ex.code, "orig": str(ex), "statement": ex.statement}
        ),
        500,
    )


@sqlBP.errorhandler(ScheduleNotFoundError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline for data source '{ex.namespace}' has not run yet. Please set up your connection.",
            }
        ),
        404,
    )


@sqlBP.errorhandler(ParseError)
def _handle(ex):
    return (jsonify({"error": True, "code": str(ex)}), 400)


@sqlBP.errorhandler(EmptyQuery)
def _handle(ex):
    return ("", 204)


def default(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serialize-able")


@sqlBP.route("/", methods=["GET"])
def index():
    return jsonify({"result": True})


@sqlBP.route("/get/<path:namespace>/<topic_name>/<design_name>", methods=["POST"])
def get_sql(namespace, topic_name, design_name):
    sqlHelper = SqlHelper()
    m5oc = sqlHelper.get_m5oc_topic(namespace, topic_name)
    design = m5oc.design(design_name)
    incoming_json = request.get_json()
    sql_dict = sqlHelper.get_sql(design, incoming_json)

    outgoing_sql = sql_dict["sql"]
    aggregates = sql_dict["aggregates"]
    query_attributes = sql_dict["query_attributes"]

    base_dict = {"sql": outgoing_sql, "error": False}
    base_dict["query_attributes"] = query_attributes
    base_dict["aggregates"] = aggregates

    if not incoming_json["run"]:
        return jsonify(base_dict)

    # we need to find the pipeline that loaded the data for this model
    # this is running off the assumption that there is only one pipeline
    # that can load data for a specific model
    project = Project.find()
    schedule_service = ScheduleService(project)
    schedule = schedule_service.find_namespace_schedule(
        m5oc.content["plugin_namespace"]
    )

    results = sqlHelper.get_query_results(
        schedule.extractor, schedule.loader, schedule.transform, outgoing_sql
    )
    base_dict["results"] = results
    base_dict["empty"] = len(results) == 0

    return jsonify(base_dict)


@sqlBP.route("/get/filter-options", methods=["GET"])
def get_filter_options():
    return jsonify(FilterOptions)



import os
import json

from meltano.core.utils import slugify

from .m5o_collection_parser import M5oCollectionParser, M5oCollectionParserTypes
from .m5o_file_parser import MeltanoAnalysisFileParser


class DashboardAlreadyExistsError(Exception):
    """Occurs when a dashboard already exists."""

    def __init__(self, dashboard, field):
        self.dashboard = dashboard
        self.field = field

    @property
    def record(self):
        return self.dashboard


class DashboardDoesNotExistError(Exception):
    """Occurs when a dashboard does not exist."""

    def __init__(self, dashboard):
        self.dashboard = dashboard


class DashboardsService:
    VERSION = "1.0.0"

    def __init__(self, project):
        self.project = project

    def get_dashboards(self):
        dashboardsParser = M5oCollectionParser(
            self.project.analyze_dir("dashboards"), M5oCollectionParserTypes.Dashboard
        )

        return dashboardsParser.parse()

    def get_dashboard(self, dashboard_id):
        dashboards = self.get_dashboards()
        dashboard = next(filter(lambda r: r["id"] == dashboard_id, dashboards), None)
        return dashboard

    def get_dashboard_by_name(self, name):
        dashboards = self.get_dashboards()
        dashboard = next(filter(lambda r: r["name"] == name, dashboards), None)
        return dashboard

    def save_dashboard(self, data):
        if "id" in data:
            existing_dashboard = self.get_dashboard(data["id"])
            if existing_dashboard:
                raise DashboardAlreadyExistsError(existing_dashboard, "id")

        name = data["name"]
        slug = slugify(name)
        file_path = self.project.analyze_dir("dashboards", f"{slug}.dashboard.m5o")

        if os.path.exists(file_path):
            with file_path.open() as f:
                existing_dashboard = json.load(f)
            raise DashboardAlreadyExistsError(existing_dashboard, "slug")

        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(
            file_path.relative_to(self.project.root), slug, data
        )
        data["version"] = DashboardsService.VERSION
        data["description"] = data["description"] or ""
        data["report_ids"] = []

        with self.project.file_update():
            with file_path.open("w") as f:
                json.dump(data, f)

        return data

    def delete_dashboard(self, data):
        dashboard = self.get_dashboard(data["id"])
        slug = dashboard["slug"]
        file_path = self.project.analyze_dir("dashboards", f"{slug}.dashboard.m5o")
        if os.path.exists(file_path):
            with self.project.file_update():
                os.remove(file_path)
        else:
            raise DashboardDoesNotExistError(data)

        return data

    def update_dashboard(self, data):
        dashboard = self.get_dashboard(data["dashboard"]["id"])
        slug = dashboard["slug"]

        file_path = self.project.analyze_dir("dashboards", f"{slug}.dashboard.m5o")
        if not os.path.exists(file_path):
            raise DashboardDoesNotExistError(data)

        new_settings = data["new_settings"]
        new_name = new_settings["name"]
        new_slug = slugify(new_name)
        new_file_path = self.project.analyze_dir(
            "dashboards", f"{new_slug}.dashboard.m5o"
        )
        is_same_file = new_slug == slug
        if not is_same_file and os.path.exists(new_file_path):
            with new_file_path.open() as f:
                existing_dashboard = json.load(f)
            raise DashboardAlreadyExistsError(existing_dashboard, "slug")

        with self.project.file_update():
            os.remove(file_path)

        dashboard["slug"] = new_slug
        dashboard["name"] = new_name
        dashboard["description"] = new_settings["description"]
        dashboard["path"] = str(new_file_path.relative_to(self.project.root))

        if "report_ids" in new_settings:
            dashboard["report_ids"] = new_settings["report_ids"]

        with self.project.file_update():
            with new_file_path.open("w") as f:
                json.dump(dashboard, f)

        return dashboard

    def add_report_to_dashboard(self, data):
        dashboard = self.get_dashboard(data["dashboard_id"])

        if data["report_id"] not in dashboard["report_ids"]:
            dashboard["report_ids"].append(data["report_id"])
            file_path = self.project.analyze_dir(
                "dashboards", f"{dashboard['slug']}.dashboard.m5o"
            )
            with self.project.file_update():
                with file_path.open("w") as f:
                    json.dump(dashboard, f)

        return dashboard

    def remove_report_from_dashboard(self, data):
        dashboard = self.get_dashboard(data["dashboard_id"])

        if data["report_id"] in dashboard["report_ids"]:
            dashboard["report_ids"].remove(data["report_id"])
            file_path = self.project.analyze_dir(
                "dashboards", f"{dashboard['slug']}.dashboard.m5o"
            )

            with self.project.file_update():
                with file_path.open("w") as f:
                    json.dump(dashboard, f)

        return dashboard

    def remove_report_from_dashboards(self, report_id):
        for dashboard in self.get_dashboards():
            self.remove_report_from_dashboard(
                {"dashboard_id": dashboard["id"], "report_id": report_id}
            )



import os
import yaml
import logging
from copy import deepcopy
from typing import Dict, List, Optional, Iterable

import meltano.core.bundle as bundle
from meltano.core.utils import nest, NotFound
from .project import Project
from .setting_definition import SettingDefinition
from .plugin import PluginDefinition, ProjectPlugin, PluginType, PluginRef, Profile
from .plugin.factory import plugin_factory
from .plugin.error import PluginMissingError


class PluginAlreadyAddedException(Exception):
    def __init__(self, plugin: PluginRef):
        self.plugin = plugin
        super().__init__()


class ConfigService:
    def __init__(self, project: Project):
        self.project = project

        self._settings = None
        self._current_config = None

    @property
    def settings(self):
        if self._settings is None:
            with bundle.find("settings.yml").open() as settings_yaml:
                settings = yaml.safe_load(settings_yaml)
            self._settings = list(map(SettingDefinition.parse, settings["settings"]))

        return self._settings

    @property
    def current_config(self):
        if self._current_config is None:
            self._current_config = self.project.meltano.extras
        return self._current_config

    def update_config(self, config):
        with self.project.meltano_update() as meltano:
            meltano.extras = config

        self._current_config = None

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    def add_to_file(self, plugin: ProjectPlugin):
        plugin = plugin_factory(plugin.type, plugin.canonical())

        if not plugin.should_add_to_file(self.project):
            return plugin

        with self.project.meltano_update() as meltano_yml:
            if plugin in self.plugins():
                raise PluginAlreadyAddedException(plugin)

            if not plugin.type in meltano_yml.plugins:
                meltano_yml.plugins[plugin.type] = []

            meltano_yml.plugins[plugin.type].append(plugin)

        return plugin

    def has_plugin(self, plugin_name: str):
        try:
            self.find_plugin(plugin_name)
            return True
        except PluginMissingError:
            return False

    def find_plugin(
        self,
        plugin_name: str,
        plugin_type: Optional[PluginType] = None,
        invokable=None,
        configurable=None,
    ) -> ProjectPlugin:
        name, profile_name = PluginRef.parse_name(plugin_name)
        try:
            plugin = next(
                plugin
                for plugin in self.plugins()
                if (
                    plugin.name == name
                    and (plugin_type is None or plugin.type == plugin_type)
                    and (invokable is None or plugin.is_invokable() == invokable)
                    and (
                        configurable is None or plugin.is_configurable() == configurable
                    )
                )
            )
            plugin.use_profile(profile_name)

            return plugin
        except StopIteration as stop:
            raise PluginMissingError(name) from stop

    def get_plugin(self, plugin_ref: PluginRef) -> ProjectPlugin:
        try:
            plugin = next(plugin for plugin in self.plugins() if plugin == plugin_ref)
            plugin.use_profile(plugin_ref.current_profile_name)

            return plugin
        except StopIteration as stop:
            raise PluginMissingError(plugin_ref.name) from stop

    def get_plugins_of_type(self, plugin_type):
        return self.project.meltano.plugins[plugin_type]

    def plugins_by_type(self):
        return {
            plugin_type: self.get_plugins_of_type(plugin_type)
            for plugin_type in PluginType
        }

    def plugins(self) -> Iterable[ProjectPlugin]:
        yield from (
            plugin
            for plugin_type, plugins in self.plugins_by_type().items()
            for plugin in plugins
        )

    def get_extractors(self):
        return self.get_plugins_of_type(PluginType.EXTRACTORS)

    def get_loaders(self):
        return self.get_plugins_of_type(PluginType.LOADERS)

    def get_transforms(self):
        return self.get_plugins_of_type(PluginType.TRANSFORMS)

    def get_models(self):
        return self.get_plugins_of_type(PluginType.MODELS)

    def get_dashboards(self):
        return self.get_plugins_of_type(PluginType.DASHBOARDS)

    def get_transformers(self):
        return self.get_plugins_of_type(PluginType.TRANSFORMERS)

    def get_files(self):
        return self.get_plugins_of_type(PluginType.FILES)

    def update_plugin(self, plugin: ProjectPlugin):
        with self.project.meltano_update() as meltano:
            # find the proper plugin to update
            idx, outdated = next(
                (i, it)
                for i, it in enumerate(meltano.plugins[plugin.type])
                if it == plugin
            )

            meltano.plugins[plugin.type][idx] = plugin

            return outdated



from copy import deepcopy
from typing import Iterable, Dict, List

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.setting_definition import SettingDefinition
from meltano.core.settings_service import (
    SettingsService,
    SettingMissingError,
    SettingValueStore,
    REDACTED_VALUE,
)
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin import (
    PluginRef,
    PluginType,
    PluginDefinition,
    ProjectPlugin,
    Profile,
)
from meltano.core.plugin.error import PluginMissingError


class PluginSettingsService(SettingsService):
    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        *args,
        plugin_discovery_service: PluginDiscoveryService = None,
        **kwargs,
    ):
        super().__init__(project, *args, **kwargs)

        self.plugin = plugin

        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            self.project, config_service=self.config_service
        )
        self._plugin_def = None

        project_settings_service = ProjectSettingsService(
            self.project, config_service=self.config_service
        )

        self.env_override = {
            **project_settings_service.env,
            **project_settings_service.as_env(),
            **self.env_override,
            **self.plugin_def.info_env,
            **self.plugin.info_env,
        }

    @property
    def plugin_def(self):
        if self._plugin_def is None:
            self._plugin_def = self.discovery_service.get_definition(self.plugin)

        return self._plugin_def

    @property
    def label(self):
        return f"{self.plugin.type.descriptor} '{self.plugin.name}'"

    @property
    def docs_url(self):
        return self.plugin_def.docs

    @property
    def _env_prefixes(self):
        return [self.plugin.name, self.plugin_def.namespace]

    @property
    def _generic_env_prefix(self):
        return f"meltano_{self.plugin.type.verb}"

    @property
    def _db_namespace(self):
        return self.plugin.qualified_name

    @property
    def extra_setting_definitions(self):
        extra_settings = deepcopy(self.plugin.extra_settings)

        # Set defaults from plugin definition
        defaults = {f"_{k}": v for k, v in self.plugin_def.all_extras.items()}

        if defaults:
            for setting in extra_settings:
                default_value = defaults.get(setting.name)
                if default_value is not None:
                    setting.value = default_value

            # Create setting definitions for unknown defaults,
            # including flattened keys of default nested object items
            extra_settings.extend(
                SettingDefinition.from_missing(
                    extra_settings, defaults, custom=False, default=True
                )
            )

        return extra_settings

    @property
    def _definitions(self):
        return [*self.plugin_def.settings, *self.extra_setting_definitions]

    @property
    def _meltano_yml_config(self):
        return {
            **self.plugin.current_config,
            **{f"_{k}": v for k, v in self.plugin.current_extras.items()},
        }

    def _update_meltano_yml_config(self, config_with_extras):
        config = self.plugin.current_config
        extras = self.plugin.current_extras

        config.clear()
        extras.clear()

        for k, v in config_with_extras.items():
            if k.startswith("_"):
                extras[k[1:]] = v
            else:
                config[k] = v

        self.config_service.update_plugin(self.plugin)

    def _process_config(self, config):
        return self.plugin.process_config(config)

    def profile_with_config(self, profile: Profile, extras=False, **kwargs):
        self.plugin.use_profile(profile)

        config_dict = {}
        config_metadata = {}

        config_with_metadata = self.config_with_metadata(extras=extras, **kwargs)
        for key, metadata in config_with_metadata.items():
            config_dict[key] = metadata.pop("value")

            metadata.pop("setting", None)

            config_metadata[key] = metadata

        return {
            **profile.canonical(),
            "config": config_dict,
            "config_metadata": config_metadata,
        }

    def profiles_with_config(self, **kwargs) -> List[Dict]:
        return [
            self.profile_with_config(profile, **kwargs)
            for profile in (Profile.DEFAULT, *self.plugin.profiles)
        ]



import json
import logging
import subprocess
import shutil
from typing import Dict
from jsonschema import Draft4Validator
from pathlib import Path
from hashlib import sha1

from meltano.core.setting_definition import SettingDefinition
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin_invoker import InvokerError
from meltano.core.job import Payload, JobFinder
from meltano.core.utils import file_has_data, truthy, flatten, merge

from . import SingerPlugin, PluginType
from .catalog import (
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    property_breadcrumb,
    select_metadata_rules,
    select_filter_metadata_rules,
)

logger = logging.getLogger(__name__)


def config_metadata_rules(config):
    flat_config = flatten(config, "dot")

    rules = []
    for key, value in flat_config.items():
        # <tap_stream_id>.<key>
        # <tap_stream_id>.<prop>.<key>
        # <tap_stream_id>.<prop>.<subprop>.<key>
        # <tap_stream_id>.properties.<prop>.<key>
        # <tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        tap_stream_id, *props, key = key.split(".")

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=key,
                value=value,
            )
        )

    return rules


def config_schema_rules(config):
    return [
        SchemaRule(
            tap_stream_id=tap_stream_id,
            breadcrumb=["properties", prop],
            payload=payload,
        )
        for tap_stream_id, stream_config in config.items()
        for prop, payload in stream_config.items()
    ]


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    @property
    def extra_settings(self):
        return [
            SettingDefinition(name="_catalog"),
            SettingDefinition(name="_state"),
            SettingDefinition(
                name="_load_schema", value="$MELTANO_EXTRACTOR_NAMESPACE"
            ),
            SettingDefinition(name="_select", kind="array", value=["*.*"]),
            SettingDefinition(
                name="_metadata",
                aliases=["metadata"],
                kind="object",
                value={},
                value_processor="nest_object",
            ),
            SettingDefinition(
                name="_schema", kind="object", value={}, value_processor="nest_object"
            ),
            SettingDefinition(name="_select_filter", kind="array", value=[]),
        ]

    def exec_args(self, plugin_invoker):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            elif "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]
            else:
                logger.warn(
                    "A catalog file was found, but it will be ignored as the extractor does not advertise the `catalog` or `properties` capability"
                )

        state_path = plugin_invoker.files["state"]
        if file_has_data(state_path):
            if "state" in plugin_invoker.capabilities:
                args += ["--state", state_path]
            else:
                logger.warn(
                    "A state file was found, but it will be ignored as the extractor does not advertise the `state` capability"
                )

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "catalog_cache_key": "tap.properties.cache_key",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke")
    def look_up_state_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.look_up_state(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def look_up_state(self, plugin_invoker, exec_args=[]):
        if "state" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support incremental state"
            )

        state_path = plugin_invoker.files["state"]

        try:
            # Delete state left over from different pipeline run for same extractor
            state_path.unlink()
        except FileNotFoundError:
            pass

        elt_context = plugin_invoker.context
        if not elt_context or not elt_context.job:
            # Running outside pipeline context: incremental state could not be loaded
            return

        if elt_context.full_refresh:
            logger.info(
                "Performing full refresh, ignoring state left behind by any previous runs."
            )
            return

        custom_state_filename = plugin_invoker.plugin_config_extras["_state"]
        if custom_state_filename:
            custom_state_path = plugin_invoker.project.root.joinpath(
                custom_state_filename
            )

            try:
                shutil.copy(custom_state_path, state_path)
                logger.info(f"Found state in {custom_state_filename}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find state file {custom_state_path}"
                ) from err

            return

        # the `state.json` is stored in the database
        state = {}
        incomplete_since = None
        finder = JobFinder(elt_context.job.job_id)

        state_job = finder.latest_with_payload(elt_context.session, flags=Payload.STATE)
        if state_job:
            logger.info(f"Found state from {state_job.started_at}.")
            incomplete_since = state_job.ended_at
            if "singer_state" in state_job.payload:
                merge(state_job.payload["singer_state"], state)

        incomplete_state_jobs = finder.with_payload(
            elt_context.session, flags=Payload.INCOMPLETE_STATE, since=incomplete_since
        )
        for state_job in incomplete_state_jobs:
            logger.info(
                f"Found and merged incomplete state from {state_job.started_at}."
            )
            if "singer_state" in state_job.payload:
                merge(state_job.payload["singer_state"], state)

        if state:
            with state_path.open("w") as state_file:
                json.dump(state, state_file, indent=2)
        else:
            logger.warning("No state was found, complete import.")

    @hook("before_invoke")
    def discover_catalog_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.discover_catalog(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def discover_catalog(self, plugin_invoker, exec_args=[]):
        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        if catalog_path.exists():
            try:
                cached_key = catalog_cache_key_path.read_text()
                new_cache_key = self.catalog_cache_key(plugin_invoker)

                if cached_key == new_cache_key:
                    logger.debug(f"Using cached catalog file")
                    return
            except FileNotFoundError:
                pass

            logging.debug("Cached catalog is outdated, running discovery...")

        # We're gonna generate a new catalog, so delete the cache key.
        try:
            catalog_cache_key_path.unlink()
        except FileNotFoundError:
            pass

        custom_catalog_filename = plugin_invoker.plugin_config_extras["_catalog"]
        if custom_catalog_filename:
            custom_catalog_path = plugin_invoker.project.root.joinpath(
                custom_catalog_filename
            )

            try:
                shutil.copy(custom_catalog_path, catalog_path)
                logger.info(f"Found catalog in {custom_catalog_path}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find catalog file {custom_catalog_path}"
                ) from err
        else:
            self.run_discovery(plugin_invoker, catalog_path)

        # test for the result to be a valid catalog
        try:
            with catalog_path.open("r") as catalog_file:
                catalog = json.load(catalog_file)
                schema_valid = Draft4Validator.check_schema(catalog)
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: invalid catalog: {err}"
            ) from err

    def run_discovery(self, plugin_invoker, catalog_path):
        if not "discover" in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support catalog discovery (the `discover` capability is not advertised)"
            )

        try:
            with catalog_path.open("w") as catalog:
                result = plugin_invoker.invoke(
                    "--discover",
                    stdout=catalog,
                    stderr=subprocess.PIPE,
                    universal_newlines=True,
                )
                stdout, stderr = result.communicate()
                exit_code = result.returncode
        except Exception:
            catalog_path.unlink()
            raise

        if exit_code != 0:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code}: {stderr.rstrip()}"
            )

    @hook("before_invoke")
    def apply_catalog_rules_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.apply_catalog_rules(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def apply_catalog_rules(self, plugin_invoker, exec_args=[]):
        if (
            not "catalog" in plugin_invoker.capabilities
            and not "properties" in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection or catalog metadata and schema rules"
            )

        config = plugin_invoker.plugin_config_extras

        schema_rules = []
        metadata_rules = []

        # If a custom catalog is provided, don't apply catalog rules
        if not config["_catalog"]:
            schema_rules.extend(config_schema_rules(config["_schema"]))

            metadata_rules.extend(select_metadata_rules(["!*.*"]))
            metadata_rules.extend(select_metadata_rules(config["_select"]))
            metadata_rules.extend(config_metadata_rules(config["_metadata"]))

        # Always apply select filters (`meltano elt` `--select` and `--exclude` options)
        metadata_rules.extend(select_filter_metadata_rules(config["_select_filter"]))

        if not schema_rules and not metadata_rules:
            return

        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        try:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            if schema_rules:
                SchemaExecutor(schema_rules).visit(catalog)

            if metadata_rules:
                MetadataExecutor(metadata_rules).visit(catalog)

            with catalog_path.open("w") as catalog_file:
                json.dump(catalog, catalog_file, indent=2)

            cache_key = self.catalog_cache_key(plugin_invoker)
            if cache_key:
                catalog_cache_key_path.write_text(cache_key)
            else:
                try:
                    catalog_cache_key_path.unlink()
                except FileNotFoundError:
                    pass
        except FileNotFoundError as err:
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is missing."
            ) from err
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is invalid: {err}"
            ) from err

    def catalog_cache_key(self, plugin_invoker):
        # If the extractor is installed as editable, don't cache because
        # the result of discovery could change at any time.
        if plugin_invoker.plugin_def.pip_url.startswith("-e"):
            return None

        extras = plugin_invoker.plugin_config_extras

        # If a custom catalog is provided, there's no need to cache
        if extras["_catalog"]:
            return None

        # The catalog should be regenerated, and the catalog cache invalidated,
        # if any settings changed that could affect discovery, or if schema or
        # metadata rules changed.
        # Changes to selection rules and selection filter rules are ignored,
        # since "selected" metadata is reset using the `!*.*` selection rule anyway.
        key_dict = {
            **plugin_invoker.plugin_config,
            "_schema": extras["_schema"],
            "_metadata": extras["_metadata"],
        }

        key_json = json.dumps(key_dict)

        return sha1(key_json.encode()).hexdigest()



import yaml
import copy

from typing import Iterable


class Canonical(object):
    """
    This class defines an object that can be reprensented as a subset of
    its attributes.

    Its purpose is to be serializable as the smallest possible form.

    The attribute rules are:
      - All attributes that are Truthy
      - All attributes that are boolean (False is valid)
      - All attributes that are listed in the `_verbatim` set and non-null
      - All attributes that start with "_" are excluded
    """

    def __init__(self, *args, **attrs):
        super(Canonical, self).__init__(*args)

        for attr, value in attrs.items():
            setattr(self, attr, value)

        self._verbatim = set()
        self._flattened = {"extras"}

    @classmethod
    def as_canonical(cls, target):
        if isinstance(target, Canonical):
            return {k: Canonical.as_canonical(v) for k, v in target}

        if isinstance(target, set):
            return list(map(Canonical.as_canonical, target))

        if isinstance(target, list):
            return list(map(Canonical.as_canonical, target))

        return copy.deepcopy(target)

    def canonical(self):
        return Canonical.as_canonical(self)

    @classmethod
    def parse(cls, obj) -> "Canonical":
        if obj is None:
            return None

        if isinstance(obj, Canonical):
            return obj

        return cls(**obj)

    def __getitem__(self, attr):
        return getattr(self, attr)

    def __setitem__(self, attr, value):
        return setattr(self, attr, value)

    def __iter__(self):
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue

            if not v:
                if k in self._verbatim:
                    if v is None:
                        continue
                else:
                    # bool values are valid and should be forwarded
                    if v is not False:
                        continue

            # empty canonicals should be skipped
            if isinstance(v, Canonical) and not dict(v):
                continue

            if k in self._flattened:
                if isinstance(v, Canonical):
                    yield from v
                else:
                    yield from v.items()
            else:
                yield (k, v)

    def __len__(self):
        return len(self.__dict__)

    def __contains__(self, obj):
        return obj in self.__dict__

    def update(self, other):
        other = Canonical.as_canonical(other)

        for k, v in other.items():
            setattr(self, k, v)

    @classmethod
    def yaml(cls, dumper, obj):
        return dumper.represent_mapping(
            "tag:yaml.org,2002:map", Canonical.as_canonical(obj), flow_style=False
        )


yaml.add_multi_representer(Canonical, Canonical.yaml)



import functools
import urllib
import click
import click.globals
import os
from pathlib import Path

from .utils import CliError

from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import pop_all
from meltano.core.migration_service import MigrationService, MigrationError
from meltano.core.db import project_engine


def database_uri_option(func):
    @click.option("--database-uri", help="System database URI")
    def decorate(*args, database_uri=None, **kwargs):
        if database_uri:
            ProjectSettingsService.config_override["database_uri"] = database_uri

        return func(*args, **kwargs)

    return functools.update_wrapper(decorate, func)


class project:
    __name__ = "project"

    def __init__(self, migrate=False):
        self.migrate = migrate

    def __call__(self, func):
        @database_uri_option
        def decorate(*args, **kwargs):
            ctx = click.globals.get_current_context()

            project = ctx.obj["project"]
            if not project:
                raise CliError(
                    f"`{ctx.command_path}` must be run inside a Meltano project."
                    "\nUse `meltano init <project_name>` to create one."
                    "\nIf you are in a project, you may be in a subfolder. Navigate to the root directory."
                )

            # register the system database connection
            settings_service = ProjectSettingsService(project)
            engine, _ = project_engine(
                project, settings_service.get("database_uri"), default=True
            )

            if self.migrate:
                try:
                    migration_service = MigrationService(engine)
                    migration_service.upgrade(silent=True)
                    migration_service.seed(project)
                except MigrationError as err:
                    raise CliError(str(err)) from err

            func(project, *args, **kwargs)

        return functools.update_wrapper(decorate, func)



from flask import url_for
from flask_security.forms import LoginForm, RegisterForm, ConfirmRegisterForm
from flask_security.utils import _datastore, get_message
from wtforms import StringField
from wtforms.validators import Required, ValidationError, Length


username_required = Required(message="USERNAME_NOT_PROVIDED")
username_validator = Length(min=6, max=32, message="USERNAME_INVALID")


def unique_username(form, field):
    if _datastore.get_user(field.data) is not None:
        msg = get_message("USERNAME_ALREADY_TAKEN", username=field.data)[0]
        raise ValidationError(msg)


class MeltanoLoginForm(LoginForm):
    email = StringField("Username or Email Address", validators=[Required()])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.next.data = self.next.data or url_for("root.bootstrap")


class UniqueUsernameMixin:
    username = StringField(
        "Username", validators=[username_required, username_validator, unique_username]
    )


class MeltanoConfirmRegisterForm(ConfirmRegisterForm, UniqueUsernameMixin):
    pass


class MeltanoRegisterFrom(RegisterForm, UniqueUsernameMixin):
    pass



import logging
from pathlib import Path

from meltano.core.db import project_engine
from meltano.core.plugin import ProjectPlugin, PluginType
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.behavior.hookable import hook
from meltano.core.venv_service import VirtualEnv
from meltano.core.setting_definition import SettingDefinition


class FilePlugin(ProjectPlugin):
    __plugin_type__ = PluginType.FILES

    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

        self._update_config = None

    @property
    def extra_settings(self):
        return [
            SettingDefinition(
                name="_update", kind="object", aliases=["update"], value={}
            )
        ]

    def is_invokable(self):
        return False

    def should_add_to_file(self, project):
        return len(self.update_config(project)) > 0

    def file_contents(self, project):
        venv = VirtualEnv(project.plugin_dir(self, "venv"))
        bundle_dir = venv.site_packages_dir.joinpath("bundle")

        return {
            path.relative_to(bundle_dir): path.read_text()
            for path in bundle_dir.glob("**/*")
            if path.is_file()
            and "__pycache__" not in path.parts
            and path != bundle_dir.joinpath("__init__.py")
        }

    def update_file_header(self, relative_path):
        return "\n".join(
            [
                f"# This file is managed by the '{self.name}' {self.type.descriptor} and updated automatically when `meltano upgrade` is run.",
                f"# To prevent any manual changes from being overwritten, remove the {self.type.descriptor} from `meltano.yml` or disable automatic updates:",
                f"#     meltano config --plugin-type={self.type} {self.name} set _update {relative_path} false",
            ]
        )

    def project_file_contents(self, project):
        def prepend_update_header(content, relative_path):
            if self.should_update_file(project, relative_path):
                content = "\n\n".join([self.update_file_header(relative_path), content])

            return content

        return {
            relative_path: prepend_update_header(content, relative_path)
            for relative_path, content in self.file_contents(project).items()
        }

    def update_config(self, project):
        if self._update_config is None:
            plugin_settings_service = PluginSettingsService(project, self)
            self._update_config = plugin_settings_service.get("_update")

        return self._update_config

    def should_update_file(self, project, relative_path):
        try:
            return self.update_config(project)[str(relative_path)]
        except KeyError:
            return False

    def file_exists(self, project, relative_path):
        return project.root_dir(relative_path).exists()

    def write_file(self, project, relative_path, content):
        project_path = project.root_dir(relative_path)
        if project_path.exists() and project_path.read_text() == content:
            return False

        project_path.parent.mkdir(parents=True, exist_ok=True)
        project_path.write_text(content)

        return True

    def write_files(self, project, files_content):
        paths = []
        for relative_path, content in files_content.items():
            if self.write_file(project, relative_path, content):
                paths.append(relative_path)

        return paths

    def files_to_create(self, project):
        def rename_if_exists(relative_path):
            exists = self.file_exists(project, relative_path)
            if not exists:
                return relative_path

            print(f"File {relative_path} already exists, keeping both versions")
            return relative_path.with_name(
                f"{relative_path.stem} ({self.name}){relative_path.suffix}"
            )

        return {
            rename_if_exists(relative_path): content
            for relative_path, content in self.project_file_contents(project).items()
        }

    def files_to_update(self, project):
        return {
            relative_path: content
            for relative_path, content in self.project_file_contents(project).items()
            if self.should_update_file(project, relative_path)
        }

    def create_files(self, project):
        return self.write_files(project, self.files_to_create(project))

    def update_files(self, project):
        return self.write_files(project, self.files_to_update(project))

    @hook("after_install")
    def after_install(self, project, reason):
        if reason is PluginInstallReason.ADD:
            print(f"Adding '{self.name}' files to project...")

            for path in self.create_files(project):
                print(f"Created {path}")
        elif reason is PluginInstallReason.UPGRADE:
            print(f"Updating '{self.name}' files in project...")

            updated_paths = self.update_files(project)
            if not updated_paths:
                print("Nothing to update")
                return

            for path in updated_paths:
                print(f"Updated {path}")
        else:
            print(
                f"Run `meltano upgrade files` to update your project's '{self.name}' files."
            )



import yaml
import fnmatch
import copy
from typing import Dict, Union
from collections import namedtuple
from enum import Enum
from typing import Optional, Iterable

from meltano.core.setting_definition import SettingDefinition
from meltano.core.behavior.hookable import HookObject
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior import NameEq
from meltano.core.utils import compact, find_named, NotFound, flatten


class VariantNotFoundError(Exception):
    def __init__(self, plugin: "PluginDefinition", variant_name):
        self.plugin = plugin
        self.variant_name = variant_name

        message = f"{plugin.type.descriptor.capitalize()} '{plugin.name}' variant '{variant_name}' is not known to Meltano. "
        message += f"Variants: {plugin.list_variant_names()}"

        super().__init__(message)


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class Profile(NameEq, Canonical):
    def __init__(self, name: str = None, label: str = None, config={}, **extras):
        super().__init__(name=name, label=label, config=config, extras=extras)


Profile.DEFAULT = Profile(name="default", label="Default")


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMS = "transforms"
    MODELS = "models"
    DASHBOARDS = "dashboards"
    ORCHESTRATORS = "orchestrators"
    TRANSFORMERS = "transformers"
    FILES = "files"

    def __str__(self):
        return self.value

    @property
    def descriptor(self):
        if self is self.__class__.FILES:
            return "file bundle"

        return self.singular

    @property
    def singular(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        return self.value[:-1]

    @property
    def verb(self):
        if self is self.__class__.TRANSFORMS:
            return self.singular

        return self.value[:-3]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def cli_arguments(cls):
        args = [type.singular for type in cls]
        args.extend([type for type in cls])
        return args

    @classmethod
    def from_cli_argument(cls, value):
        if not value.endswith("s"):
            value += "s"

        return cls(value)


class PluginRef:
    def __init__(self, plugin_type: Union[str, PluginType], name: str):
        self._type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )
        self.name, self._current_profile_name = self.parse_name(name)

    @classmethod
    def parse_name(cls, name: str):
        name, *profile_name = name.split("@")
        profile_name = next(iter(profile_name), Profile.DEFAULT.name)

        return (name, profile_name)

    @property
    def type(self):
        return self._type

    @property
    def current_profile_name(self):
        return self._current_profile_name

    @property
    def full_name(self):
        return f"{self.name}@{self.current_profile_name}"

    @property
    def qualified_name(self):
        parts = (self.type, self.name, self.current_profile_name)

        return ".".join(compact(parts))

    @property
    def info(self):
        return {"name": self.name, "profile": self.current_profile_name}

    @property
    def info_env(self):
        # MELTANO_EXTRACTOR_...
        return flatten({"meltano": {self.type.singular: self.info}}, "env_var")

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __hash__(self):
        return hash((self.type, self.name, self.current_profile_name))


class ProjectPlugin(HookObject, Canonical, PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: Optional[str] = None,
        custom_definition: Optional["PluginDefinition"] = None,
        variant: Optional[str] = None,
        pip_url: Optional[str] = None,
        config: Optional[dict] = {},
        profiles: Optional[list] = [],
        **extras,
    ):
        if namespace:
            custom_definition = PluginDefinition(
                plugin_type, name, namespace, variant=variant, pip_url=pip_url, **extras
            )
            extras = {}

        if custom_definition:
            # Any properties considered "extra" by the embedded plugin definition
            # should be considered extras of the project plugin, since they are
            # the current values, not default values.
            extras = {**custom_definition.extras, **extras}
            custom_definition.extras = {}

        super().__init__(
            plugin_type,
            name,
            # Attributes will be listed in meltano.yml in this order:
            custom_definition=custom_definition,
            variant=variant,
            pip_url=pip_url,
            config=copy.deepcopy(config),
            extras=extras,
            profiles=list(map(Profile.parse, profiles)),
        )

        self._flattened.add("custom_definition")

    def is_installable(self):
        return self.pip_url is not None

    def is_invokable(self):
        return self.is_installable()

    def is_configurable(self):
        return True

    def should_add_to_file(self, project):
        return True

    @property
    def runner(self):
        return None

    @property
    def extra_settings(self):
        return []

    def is_custom(self):
        return self.custom_definition is not None

    def get_profile(self, profile_name: str) -> Profile:
        if profile_name == Profile.DEFAULT.name:
            return Profile.DEFAULT

        return find_named(self.profiles, profile_name)

    def use_profile(self, profile_or_name: Union[str, Profile]):
        if profile_or_name is None:
            profile = Profile.DEFAULT
        elif isinstance(profile_or_name, Profile):
            profile = profile_or_name
        else:
            profile = self.get_profile(profile_or_name)

        self._current_profile_name = profile.name

    @property
    def current_profile(self):
        return self.get_profile(self.current_profile_name)

    @property
    def current_config(self):
        return (
            self.config
            if self.current_profile is Profile.DEFAULT
            else self.current_profile.config
        )

    @property
    def current_extras(self):
        return (
            self.extras
            if self.current_profile is Profile.DEFAULT
            else self.current_profile.extras
        )

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    def add_select_filter(self, filter: str):
        select = self.extras.get("select", [])
        select.append(filter)
        self.extras["select"] = select

    def add_profile(self, name: str, config: dict = None, label: str = None):
        profile = Profile(name=name, config=config, label=label)

        self.profiles.append(profile)
        return profile

    def process_config(self, config):
        return config


class Variant(NameEq, Canonical):
    ORIGINAL_NAME = "original"

    def __init__(
        self,
        name: str = None,
        original: Optional[bool] = None,
        deprecated: Optional[bool] = None,
        docs: Optional[str] = None,
        repo: Optional[str] = None,
        pip_url: Optional[str] = None,
        executable: Optional[str] = None,
        capabilities: Optional[list] = [],
        settings_group_validation: Optional[list] = [],
        settings: Optional[list] = [],
        **extras,
    ):
        super().__init__(
            name=name,
            original=original,
            deprecated=deprecated,
            docs=docs,
            repo=repo,
            pip_url=pip_url,
            executable=executable,
            capabilities=list(capabilities),
            settings_group_validation=list(settings_group_validation),
            settings=list(map(SettingDefinition.parse, settings)),
            extras=extras,
        )


class PluginDefinition(Canonical, PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        hidden: Optional[bool] = None,
        label: Optional[str] = None,
        logo_url: Optional[str] = None,
        description: Optional[str] = None,
        variant: Optional[str] = None,
        variants: Optional[list] = [],
        **extras,
    ):
        if not variants:
            variant = Variant(variant, **extras)

            # Any properties considered "extra" by the variant should be
            # considered extras of the plugin definition.
            extras = variant.extras
            variant.extras = {}

            variants = [variant]

        super().__init__(
            plugin_type,
            name,
            # Attributes will be listed in meltano.yml in this order:
            namespace=namespace,
            hidden=hidden,
            label=label,
            logo_url=logo_url,
            description=description,
            extras=extras,
            variants=list(map(Variant.parse, variants)),
        )

        self.use_variant(variant)

    def __iter__(self):
        for k, v in super().__iter__():
            if k == "variants" and len(v) == 1:
                # If there is only a single variant, its properties can be
                # nested in the plugin definition
                for variant_k, variant_v in v[0]:
                    if variant_k == "name":
                        variant_k = "variant"

                    yield (variant_k, variant_v)
            else:
                yield (k, v)

    @property
    def info(self):
        return {
            **super().info,
            "namespace": self.namespace,
            "variant": self.current_variant_name or Variant.ORIGINAL_NAME,
        }

    @property
    def current_variant_name(self):
        return self._current_variant_name

    def get_variant(self, variant_name: str) -> Profile:
        try:
            return find_named(self.variants, variant_name)
        except NotFound as err:
            raise VariantNotFoundError(self, variant_name) from err

    def use_variant(self, variant_or_name: Union[str, Variant] = None):
        if variant_or_name is None:
            variant = self.variants[0]
        elif isinstance(variant_or_name, Variant):
            variant = variant_or_name
        elif variant_or_name == Variant.ORIGINAL_NAME:
            try:
                variant = next(v for v in self.variants if v.original)
            except StopIteration:
                variant = self.variants[0]
        else:
            variant = self.get_variant(variant_or_name)

        self._current_variant_name = variant.name

    @property
    def current_variant(self):
        return self.get_variant(self.current_variant_name)

    def list_variant_names(self):
        names = []

        for i, variant in enumerate(self.variants):
            name = variant.name or Variant.ORIGINAL_NAME

            if i == 0:
                name += " (default)"
            elif variant.deprecated:
                name += " (deprecated)"

            names.append(name)

        return ", ".join(names)

    @property
    def all_extras(self):
        return {**self.extras, **self.current_variant.extras}

    def __getattr__(self, attr):
        return getattr(self.current_variant, attr)

    def in_project(self, custom=False) -> ProjectPlugin:
        return ProjectPlugin(
            self.type,
            self.name,
            variant=self.current_variant_name,
            pip_url=self.pip_url,
            custom_definition=(self if custom else None),
        )



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION") as version_file:
    version = version_file.read().strip()

requires = [
    "aiohttp==3.4.4",
    "alembic==1.0.11",
    "atomicwrites==1.2.1",
    "authlib==0.10",
    "async_generator==1.10",
    "backoff==1.8.0",
    "bcrypt==3.1.6",
    "cerberus==1.2",
    "click==7.0",
    "click-default-group==1.2.1",
    "colorama==0.3.9",
    "cryptography==2.9.2",
    "gunicorn==19.9.0",
    "ipython==7.5.0",
    "jsonschema==2.6.0",
    "markdown==3.0.1",
    "networkx==2.2",
    "psycopg2-binary==2.8.5",
    "psutil==5.6.3",
    "pyhumps==1.2.2",
    "pypika==0.25.1",
    "python-dotenv==0.14.0",
    "pyyaml==5.3.1",
    "smtpapi==0.4.1",
    "snowflake-sqlalchemy==1.2.3",
    "sqlalchemy==1.3.19",
    "fasteners==0.15.0",
    "flask>=1.0.2",
    "flask-cors==3.0.7",
    "flask-executor==0.9.2",
    "flask-sqlalchemy==2.4.4",
    "flask-restful==0.3.7",
    "flatten-dict==0.1.0",
    "meltano-flask-security==0.1.0",
    "pyhocon==0.3.51",
    "python-gitlab==1.8.0",
    "requests==2.23.0",
    "simplejson==3.16.0",
    "urllib3==1.23",
    "sqlparse==0.3.0",
    "watchdog==0.9.0",
    "werkzeug==0.16.1",
]

# conflicts resolution, see https://gitlab.com/meltano/meltano/issues/193
conflicts = ["aenum==2.1.2", "idna==2.7", "asn1crypto==1.3.0", "wtforms==2.2.1"]

dev_requires = [
    "asynctest==0.12.2",
    "black==18.9b0",
    "bumpversion==0.5.3",
    "changelog-cli==0.6.2",
    "coverage==4.5.4",
    "freezegun==0.3.12",
    "pytest==4.3.1",
    "pytest-asyncio==0.10.0",
    "pytest-cov==2.6.1",
    "requests-mock==1.6.0",
]

infra_requires = ["ansible==2.9.4"]

setup(
    name="meltano",
    version=version,
    author="Meltano Team & Contributors",
    author_email="meltano@gitlab.com",
    description="Meltano",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/meltano/meltano",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    tests_require=dev_requires,
    # run `make requirements.txt` after editing
    install_requires=[*conflicts, *requires],
    extras_require={"dev": dev_requires, "infra": infra_requires},
    entry_points={"console_scripts": ["meltano = meltano.cli:main"]},
)



import subprocess

from .project import Project
from .project_settings_service import ProjectSettingsService, SettingValueStore


class MeltanoInvoker:
    def __init__(self, project, settings_service: ProjectSettingsService = None):
        self.project = project
        self.settings_service = settings_service or ProjectSettingsService(project)

    def invoke(self, args, command="meltano", env={}, **kwargs):
        base_env = self.settings_service.env
        overridden_config_env = self.settings_service.as_env(
            source=SettingValueStore.CONFIG_OVERRIDE
        )

        return subprocess.run(
            [command, *args], **kwargs, env={**base_env, **overridden_config_env, **env}
        )



from . import Job, State
from meltano.core.db import DB


class JobFinder:
    """
    Query builder for the `Job` model for a certain `elt_uri`.
    """

    def __init__(self, job_id: str):
        self.job_id = job_id

    def latest(self, session):
        return (
            session.query(Job)
            .filter((Job.job_id == self.job_id))
            .order_by(Job.started_at.desc())
            .first()
        )

    def successful(self, session):
        return session.query(Job).filter(
            (Job.job_id == self.job_id)
            & (Job.state == State.SUCCESS)
            & Job.ended_at.isnot(None)
        )

    def latest_success(self, session):
        return self.successful(session).order_by(Job.ended_at.desc()).first()

    def with_payload(self, session, flags=0, since=None):
        query = (
            session.query(Job)
            .filter(
                (Job.job_id == self.job_id)
                & (Job.payload_flags != 0)
                & (Job.payload_flags.op("&")(flags) == flags)
                & Job.ended_at.isnot(None)
            )
            .order_by(Job.ended_at.asc())
        )

        if since:
            query = query.filter(Job.ended_at > since)

        return query

    def latest_with_payload(self, session, **kwargs):
        return (
            self.with_payload(session, **kwargs)
            .order_by(None)  # Reset ascending order
            .order_by(Job.ended_at.desc())
            .first()
        )



from pypika import functions as fn

from .substitution import Substitution


class TimeframePeriod:
    def __init__(self, timeframe, period, table):
        self.timeframe = timeframe
        self.period = period
        self.field = Substitution(
            timeframe["sql"], table, timeframe, alias=timeframe["name"]
        )

    @property
    def alias(self):
        return f"{self.timeframe['name']}.{self.period['label'].lower()}"

    @property
    def sql(self):
        return fn.Cast(
            fn.Extract(self.period["label"], self.field.sql),
            "INTEGER",
            alias=self.alias,
        )



import os
import subprocess
import logging
import shutil
from pathlib import Path
from typing import Union

from meltano.core.project import Project
from meltano.core.plugin import ProjectPlugin


class PluginConfigService:
    """
    Manage the plugin configuration files. Each plugin can expose a
    set of files that should be stubbed using environment variables.
    """

    def __init__(
        self,
        plugin: ProjectPlugin,
        config_dir: Union[str, Path],
        run_dir: Union[str, Path],
    ):
        self.plugin = plugin
        self.config_dir = Path(config_dir)
        self.run_dir = Path(run_dir)

    def configure(self):
        os.makedirs(self.run_dir, exist_ok=True)

        config_file = self.config_dir.joinpath
        run_file = self.run_dir.joinpath

        # grab the list of files the plugin needs
        stubs = (
            (run_file(file_name), config_file(file_name))
            for file_id, file_name in self.plugin.config_files.items()
            if file_id != "config"
        )

        stubbed = []
        # stub the files from the env
        for dst, src in stubs:
            try:
                shutil.copy(src, dst)
                stubbed.append(dst)
            except FileNotFoundError:
                logging.debug(
                    f"Could not find {src.name} in {src.resolve()}, skipping."
                )

        return stubbed



import asyncio
import json
import logging
import sqlalchemy
import shutil
from flask import request, url_for, jsonify, make_response, Response, send_file
from flask_restful import Api, Resource, fields, marshal, marshal_with
from werkzeug.exceptions import Conflict, UnprocessableEntity

from meltano.core.job import JobFinder, State
from meltano.core.behavior.canonical import Canonical
from meltano.core.plugin import PluginRef, Profile
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin.settings_service import (
    PluginSettingsService,
    SettingValueStore,
)
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.schedule_service import (
    ScheduleService,
    ScheduleAlreadyExistsError,
    ScheduleDoesNotExistError,
)
from meltano.core.utils import flatten, iso8601_datetime, slugify
from meltano.core.logging import (
    JobLoggingService,
    MissingJobLogException,
    SizeThresholdJobLogException,
)
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import block_if_readonly
from meltano.api.models import db
from meltano.api.models.subscription import Subscription
from meltano.api.json import freeze_keys
from meltano.api.executor import run_elt
from flask_security import roles_required
from .errors import InvalidFileNameError
from .upload_helper import InvalidFileTypeError, InvalidFileSizeError, UploadHelper
from .utils import enforce_secure_filename


def decorate_profile_config(profile):
    profile["config"] = freeze_keys(profile.get("config", {}))

    config_metadata = profile.get("config_metadata", {})
    for key, metadata in config_metadata.items():
        try:
            metadata["source_label"] = metadata["source"].label
            metadata["auto_store_label"] = metadata["auto_store"].label
        except KeyError:
            pass

    profile["config_metadata"] = freeze_keys(config_metadata)


def validate_plugin_config(
    plugin: PluginRef, name, value, project: Project, settings: PluginSettingsService
):
    setting_def = settings.find_setting(name)
    # we want to prevent the edition of protected settings from the UI
    if setting_def.protected:
        logging.warning("Cannot set a 'protected' setting externally.")
        return False

    if setting_def.kind == "file" and value and value != "":
        uploads_directory = project.extract_dir(plugin.full_name)
        resolved_file_path = project.root_dir(value).resolve()
        if not str(resolved_file_path).startswith(str(uploads_directory) + "/"):
            logging.warning(
                "Cannot set a file configuration to a path outside the project directory"
            )
            return False

    old_value, metadata = settings.get_with_metadata(name, session=db.session)
    if not metadata["overwritable"]:
        logging.warning("Cannot overwrite this setting.")
        return False

    return True


orchestrationsBP = APIBlueprint("orchestrations", __name__)
orchestrationsAPI = Api(
    orchestrationsBP,
    errors={
        "UnprocessableEntity": {
            "error": True,
            "code": "The subscription could not be created.",
            "status": UnprocessableEntity.code,
        },
        "Conflict": {
            "error": True,
            "code": "A subscription already exists for this address.",
            "status": Conflict.code,
        },
    },
)


@orchestrationsBP.errorhandler(ScheduleAlreadyExistsError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline with the name '{ex.schedule.name}' already exists. Try renaming the pipeline.",
            }
        ),
        409,
    )


@orchestrationsBP.errorhandler(ScheduleDoesNotExistError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"A pipeline with the name '{ex.name}' does not exist..",
            }
        ),
        404,
    )


@orchestrationsBP.errorhandler(InvalidFileNameError)
def _handle(ex):
    return (jsonify({"error": True, "code": f"The file lacks a valid name."}), 400)


@orchestrationsBP.errorhandler(InvalidFileTypeError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The file '{ex.file.filename}' must be one of the following types: {ex.extensions}",
            }
        ),
        400,
    )


@orchestrationsBP.errorhandler(InvalidFileSizeError)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"The file '{ex.file.filename}' is empty or exceeds the {ex.max_file_size} size limit.",
            }
        ),
        400,
    )


@orchestrationsBP.errorhandler(MissingJobLogException)
def _handle(ex):
    return (jsonify({"error": False, "code": str(ex)}), 204)


@orchestrationsBP.route("/jobs/state", methods=["POST"])
def job_state() -> Response:
    """
    Endpoint for getting the status of N jobs
    """
    project = Project.find()
    poll_payload = request.get_json()
    job_ids = poll_payload["job_ids"]

    jobs = []
    for job_id in job_ids:
        finder = JobFinder(job_id)
        state_job = finder.latest(db.session)
        # Validate existence first as a job may not be queued yet as a result of
        # another prerequisite async process (dbt installation for example)
        if state_job:
            state_job_success = finder.latest_success(db.session)
            jobs.append(
                {
                    "job_id": job_id,
                    "is_complete": state_job.is_complete(),
                    "has_error": state_job.has_error(),
                    "started_at": state_job.started_at,
                    "ended_at": state_job.ended_at,
                    "has_ever_succeeded": state_job_success.is_success()
                    if state_job_success
                    else None,
                }
            )

    return jsonify({"jobs": jobs})


@orchestrationsBP.route("/jobs/<job_id>/log", methods=["GET"])
def job_log(job_id) -> Response:
    """
    Endpoint for getting the most recent log generated by a job with job_id
    """
    project = Project.find()
    try:
        log_service = JobLoggingService(project)
        log = log_service.get_latest_log(job_id)
        has_log_exceeded_max_size = False
    except SizeThresholdJobLogException as err:
        log = None
        has_log_exceeded_max_size = True

    finder = JobFinder(job_id)
    state_job = finder.latest(db.session)
    state_job_success = finder.latest_success(db.session)

    return jsonify(
        {
            "job_id": job_id,
            "log": log,
            "has_log_exceeded_max_size": has_log_exceeded_max_size,
            "has_error": state_job.has_error() if state_job else False,
            "started_at": state_job.started_at if state_job else None,
            "ended_at": state_job.ended_at if state_job else None,
            "trigger": state_job.trigger if state_job else None,
            "has_ever_succeeded": state_job_success.is_success()
            if state_job_success
            else None,
        }
    )


@orchestrationsBP.route("/jobs/<job_id>/download", methods=["GET"])
def download_job_log(job_id) -> Response:
    """
    Endpoint for downloading a job log with job_id
    """
    project = Project.find()
    log_service = JobLoggingService(project)
    return send_file(log_service.get_downloadable_log(job_id), mimetype="text/plain")


@orchestrationsBP.route("/run", methods=["POST"])
def run():
    project = Project.find()
    schedule_payload = request.get_json()
    job_id = run_elt(project, schedule_payload)

    return jsonify({"job_id": job_id}), 202


@orchestrationsBP.route(
    "/<plugin_ref:plugin_ref>/configuration/upload-file", methods=["POST"]
)
@block_if_readonly
def upload_plugin_configuration_file(plugin_ref) -> Response:
    """
    Endpoint for uploading a file for a specific plugin's configuration profile
    """

    file = request.files["file"]
    setting_name = enforce_secure_filename(request.form["setting_name"])
    tmp = request.form.get("tmp", False)

    project = Project.find()
    directory = project.extract_dir(
        plugin_ref.full_name, ("tmp" if tmp else ""), setting_name
    )
    upload_helper = UploadHelper()
    file_path = upload_helper.upload_file(directory, file)

    return jsonify({"path": file_path, "setting_name": setting_name}), 200


@orchestrationsBP.route(
    "/<plugin_ref:plugin_ref>/configuration/delete-uploaded-file", methods=["POST"]
)
@block_if_readonly
def delete_plugin_configuration_file(plugin_ref) -> Response:
    """
    Endpoint for deleting a file for a specific plugin's configuration profile
    """

    payload = request.get_json()
    setting_name = enforce_secure_filename(payload["setting_name"])
    tmp = payload.get("tmp", False)

    project = Project.find()
    directory = project.extract_dir(
        plugin_ref.full_name, ("tmp" if tmp else ""), setting_name
    )
    shutil.rmtree(directory)

    return jsonify({"setting_name": setting_name}), 200


@orchestrationsBP.route("/<plugin_ref:plugin_ref>/configuration", methods=["GET"])
def get_plugin_configuration(plugin_ref) -> Response:
    """
    Endpoint for getting a plugin's configuration profiles
    """

    project = Project.find()
    config_service = ConfigService(project)
    plugin = config_service.get_plugin(plugin_ref)

    discovery_service = PluginDiscoveryService(project, config_service=config_service)
    settings = PluginSettingsService(
        project,
        plugin,
        config_service=config_service,
        plugin_discovery_service=discovery_service,
        show_hidden=False,
    )

    try:
        plugin_def = discovery_service.get_definition(plugin)
        settings_group_validation = plugin_def.settings_group_validation
    except PluginNotFoundError:
        settings_group_validation = []

    profiles = settings.profiles_with_config(redacted=True, session=db.session)
    for profile in profiles:
        decorate_profile_config(profile)

    return jsonify(
        {
            "profiles": profiles,
            "settings": Canonical.as_canonical(settings.definitions(extras=False)),
            "settings_group_validation": settings_group_validation,
        }
    )


@orchestrationsBP.route(
    "/<plugin_ref:plugin_ref>/configuration/profiles", methods=["POST"]
)
@block_if_readonly
def add_plugin_configuration_profile(plugin_ref) -> Response:
    """
    Endpoint for adding a configuration profile to a plugin
    """
    payload = request.get_json()
    project = Project.find()
    config = ConfigService(project)
    plugin = config.get_plugin(plugin_ref)
    settings = PluginSettingsService(project, plugin, config_service=config)

    # create the new profile for this plugin
    name = payload["name"]
    profile = plugin.add_profile(slugify(name), label=name)

    config.update_plugin(plugin)

    profile_config = settings.profile_with_config(
        profile, redacted=True, session=db.session
    )
    decorate_profile_config(profile_config)

    return jsonify(profile_config)


@orchestrationsBP.route("/<plugin_ref:plugin_ref>/configuration", methods=["PUT"])
@block_if_readonly
def save_plugin_configuration(plugin_ref) -> Response:
    """
    Endpoint for persisting a plugin configuration
    """
    project = Project.find()
    payload = request.get_json()
    config_service = ConfigService(project)
    plugin = config_service.get_plugin(plugin_ref)

    # TODO iterate pipelines and save each, also set this connector's profile (reuse `pipelineInFocusIndex`?)

    settings = PluginSettingsService(
        project, plugin, config_service=config_service, show_hidden=False
    )

    for profile in payload:
        # select the correct profile
        name = profile["name"]
        plugin.use_profile(plugin.get_profile(name))

        for name, value in profile["config"].items():
            if not validate_plugin_config(plugin, name, value, project, settings):
                continue

            if value == "":
                settings.unset(name, session=db.session)
            else:
                settings.set(name, value, session=db.session)

    profiles = settings.profiles_with_config(redacted=True, session=db.session)
    for profile in profiles:
        decorate_profile_config(profile)

    return jsonify(profiles)


@orchestrationsBP.route("/<plugin_ref:plugin_ref>/configuration/test", methods=["POST"])
@block_if_readonly
def test_plugin_configuration(plugin_ref) -> Response:
    """
    Endpoint for testing a plugin configuration's valid connection
    """
    project = Project.find()
    payload = request.get_json()
    config_service = ConfigService(project)
    plugin = config_service.get_plugin(plugin_ref)

    # load the correct profile
    plugin.use_profile(plugin.get_profile(payload.get("profile")))

    settings = PluginSettingsService(
        project, plugin, config_service=config_service, show_hidden=False
    )

    config = payload.get("config", {})
    valid_config = {
        name: value
        for name, value in config.items()
        if validate_plugin_config(plugin, name, value, project, settings)
    }
    settings.config_override = PluginSettingsService.unredact(valid_config)

    async def test_stream(tap_stream) -> bool:
        while not tap_stream.at_eof():
            message = await tap_stream.readline()
            json_dict = json.loads(message)
            if json_dict["type"] == "RECORD":
                return True

        return False

    async def test_extractor():
        process = None
        try:
            invoker = invoker_factory(project, plugin, plugin_settings_service=settings)
            with invoker.prepared(db.session):
                process = await invoker.invoke_async(stdout=asyncio.subprocess.PIPE)
                return await test_stream(process.stdout)
        except Exception as err:
            logging.debug(err)
            # if anything happens, this is not successful
            return False
        finally:
            try:
                if process:
                    psutil.Process(process.pid).terminate()
            except Exception as err:
                logging.debug(err)

    loop = asyncio.get_event_loop()
    success = loop.run_until_complete(test_extractor())

    return jsonify({"is_success": success}), 200


@orchestrationsBP.route("/pipeline-schedules", methods=["GET"])
def get_pipeline_schedules():
    """
    Endpoint for getting the pipeline schedules
    """
    project = Project.find()
    schedule_service = ScheduleService(project)
    schedules = list(map(dict, schedule_service.schedules()))
    for schedule in schedules:
        finder = JobFinder(schedule["name"])
        state_job = finder.latest(db.session)
        schedule["has_error"] = state_job.has_error() if state_job else False
        schedule["is_running"] = state_job.is_running() if state_job else False
        schedule["job_id"] = schedule["name"]
        schedule["started_at"] = state_job.started_at if state_job else None
        schedule["ended_at"] = state_job.ended_at if state_job else None
        schedule["trigger"] = state_job.trigger if state_job else None

        state_job_success = finder.latest_success(db.session)
        schedule["has_ever_succeeded"] = (
            state_job_success.is_success() if state_job_success else None
        )

        schedule["start_date"] = (
            schedule["start_date"].date().isoformat()
            if schedule["start_date"]
            else None
        )

    return jsonify(schedules)


@orchestrationsBP.route("/pipeline-schedules", methods=["POST"])
@block_if_readonly
def save_pipeline_schedule() -> Response:
    """
    Endpoint for persisting a pipeline schedule
    """
    payload = request.get_json()
    # Airflow requires alphanumeric characters, dashes, dots and underscores exclusively
    name = payload["name"]
    slug = slugify(name)
    extractor = payload["extractor"]
    loader = payload["loader"]
    transform = payload["transform"]
    interval = payload["interval"]

    project = Project.find()
    schedule_service = ScheduleService(project)

    schedule = schedule_service.add(
        db.session, slug, extractor, loader, transform, interval
    )

    schedule = dict(schedule)
    schedule["start_date"] = (
        schedule["start_date"].date().isoformat() if schedule["start_date"] else None
    )

    return jsonify(schedule), 201


@orchestrationsBP.route("/pipeline-schedules", methods=["PUT"])
@block_if_readonly
def update_pipeline_schedule() -> Response:
    """
    Endpoint for updating a pipeline schedule
    """
    payload = request.get_json()
    project = Project.find()
    schedule_service = ScheduleService(project)

    interval = payload["interval"]
    plugin_namespace = payload["plugin_namespace"]
    schedule = schedule_service.find_namespace_schedule(plugin_namespace)
    schedule.interval = interval
    schedule_service.update_schedule(schedule)

    schedule = dict(schedule)
    schedule["start_date"] = (
        schedule["start_date"].date().isoformat() if schedule["start_date"] else None
    )

    return jsonify(schedule), 201


@orchestrationsBP.route("/pipeline-schedules", methods=["DELETE"])
@block_if_readonly
def delete_pipeline_schedule() -> Response:
    """
    Endpoint for deleting a pipeline schedule
    """
    payload = request.get_json()
    project = Project.find()
    schedule_service = ScheduleService(project)

    name = payload["name"]
    schedule_service.remove(name)
    return jsonify(name), 201


class SubscriptionsResource(Resource):
    SubscriptionDefinition = {
        "id": fields.String,
        "recipient": fields.String,
        "event_type": fields.String,
        "source_type": fields.String,
        "source_id": fields.String,
        "created_at": fields.DateTime,
    }

    @marshal_with(SubscriptionDefinition)
    def get(self):
        return Subscription.query.all()

    @marshal_with(SubscriptionDefinition)
    def post(self):
        payload = request.get_json()

        try:
            subscription = Subscription(**payload)
            db.session.add(subscription)
            db.session.commit()
        except AssertionError as err:
            raise UnprocessableEntity() from err
        except sqlalchemy.exc.IntegrityError:
            raise Conflict()

        return subscription, 201


class SubscriptionResource(Resource):
    def delete(self, id):
        Subscription.query.filter_by(id=id).delete()
        db.session.commit()

        return "", 204


orchestrationsAPI.add_resource(SubscriptionsResource, "/subscriptions")
orchestrationsAPI.add_resource(SubscriptionResource, "/subscriptions/<id>")
